# -----------------------------------------------------------------------
# Copyright (c) 2008-2024 Jendrik Seipp
#
# RedNotebook is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# RedNotebook is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with RedNotebook; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
# -----------------------------------------------------------------------

import codecs
import locale
import logging
import os
import platform
import subprocess
import sys

import gi


ENCODING = sys.getfilesystemencoding() or locale.getlocale()[1] or "UTF-8"
LANGUAGE = locale.getdefaultlocale()[0]
REMOTE_PROTOCOLS = ["http", "ftp", "irc"]

IS_WIN = sys.platform.startswith("win")
IS_MAC = sys.platform == "darwin"

LOCAL_FILE_PEFIX = "file:///" if IS_WIN else "file://"


try:
    gi.require_version("GIRepository", "3.0")
except ValueError:
    try:
        gi.require_version("GIRepository", "2.0")
    except ValueError:
        sys.exit("Please install GIRepository (package gir1.2-glib-* on Ubuntu).")
from gi.repository import GIRepository


repo = GIRepository.Repository()
logging.info(
    f"Available versions of the WebKit2 namespace: {repo.enumerate_versions('WebKit2')}"
)


def _is_nvidia_graphics_detected():
    """
    Detect if the system has Nvidia graphics drivers that might be affected
    by the WebKitGTK DMA-BUF renderer bug.

    Returns True if Nvidia graphics are detected, False otherwise.
    """
    # Fast path: check common file indicators and return immediately if any exist.
    if any(
        os.path.exists(p)
        for p in (
            "/proc/driver/nvidia/version",
            "/dev/nvidia0",
            "/dev/nvidiactl",
        )
    ):
        return True

    # Fallback: use lspci output (best effort, ignore errors/timeouts).
    try:
        result = subprocess.run(["lspci"], capture_output=True, text=True, timeout=5)
        if result.returncode == 0 and "nvidia" in result.stdout.lower():
            return True
    except (subprocess.TimeoutExpired, FileNotFoundError):
        pass
    return False


def _apply_webkit_nvidia_workaround():
    """
    Apply workaround for WebKitGTK DMA-BUF renderer bug with Nvidia drivers.

    This sets WEBKIT_DISABLE_DMABUF_RENDERER=1 if Nvidia graphics are detected
    and the environment variable is not already set.
    """
    env_var = "WEBKIT_DISABLE_DMABUF_RENDERER"

    # Only apply if not already set by user
    if env_var not in os.environ:
        if _is_nvidia_graphics_detected():
            os.environ[env_var] = "1"
            logging.info(
                "Nvidia graphics detected. Setting WEBKIT_DISABLE_DMABUF_RENDERER=1 "
                "to work around WebKitGTK rendering issues."
            )


# Apply Nvidia workaround before importing WebKit2
if not IS_WIN:  # Only apply on Linux/Unix systems
    _apply_webkit_nvidia_workaround()


try:
    gi.require_version("WebKit2", "4.1")
except ValueError as err:
    logging.warning(
        f"WebKit2 4.1 not found. Trying to use arbitrary version. "
        f"Error message: '{err}'"
    )

try:
    from gi.repository import WebKit2

    # Don't log the version as it leads to a warning about a failed assertion.
    # logging.info(
    #    f"Loaded version of the WebKit2 namespace: {repo.get_version('WebKit2')}"
    # )
except ImportError as err:
    logging.info("Failed to load the WebKit2 namespace")
    WebKit2 = None
    if not IS_WIN:
        logging.info(
            f"WebKit2Gtk not found. Please install"
            f" it if you want in-app previews."
            f" On Debian/Ubuntu you need the gir1.2-webkit2-4.1 package."
            f' Error message: "{err}"'
        )


def has_system_tray():
    return IS_WIN  # A smarter detection is needed here ;)


def main_is_frozen():
    return hasattr(sys, "frozen")


if main_is_frozen():
    app_dir = sys._MEIPASS  # os.path.dirname(sys.executable)
else:
    app_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

if main_is_frozen():
    locale_dir = os.path.join(app_dir, "share", "locale")
else:
    locale_dir = os.path.join(sys.prefix, "share", "locale")

image_dir = os.path.join(app_dir, "images")
frame_icon_dir = os.path.join(image_dir, "rednotebook-icon")
files_dir = os.path.join(app_dir, "files")

user_home_dir = os.path.expanduser("~")


class Filenames(dict):
    """
    Dictionary for dirnames and filenames
    """

    def __init__(self, config):
        for key, value in globals().items():
            # Exclude "get_main_dir()"
            if key.lower().endswith("dir") and isinstance(value, str):
                value = os.path.abspath(value)
                self[key] = value
                setattr(self, key, value)

        self.portable = bool(config.read("portable"))

        self.journal_user_dir = self.get_user_dir(config)

        self.data_dir = self.default_data_dir

        # Assert that all dirs and files are in place so that logging can take start
        make_directories(
            [self.journal_user_dir, self.data_dir, self.template_dir, self.temp_dir]
        )
        make_files([(self.config_file, ""), (self.log_file, "")])

        self.last_pic_dir = self.user_home_dir
        self.last_file_dir = self.user_home_dir

        self.forbidden_dirs = [user_home_dir, self.journal_user_dir]

    def get_user_dir(self, config):
        if not (custom := config.read("userDir")):
            return (
                os.path.join(self.app_dir, "user")
                if self.portable
                else os.path.join(self.user_home_dir, ".rednotebook")
            )
        # If a custom user dir has been set,
        # construct the absolute path (if not absolute already)
        # and use it
        if not os.path.isabs(custom):
            custom = os.path.join(self.app_dir, custom)
        return custom

    def is_valid_journal_path(self, path):
        return os.path.isdir(path) and os.path.abspath(path) not in self.forbidden_dirs

    def __getattribute__(self, attr):
        user_paths = {
            "template_dir": "templates",
            "temp_dir": "tmp",
            "default_data_dir": "data",
            "config_file": "configuration.cfg",
            "log_file": "rednotebook.log",
        }

        if attr in user_paths:
            return os.path.join(self.journal_user_dir, user_paths.get(attr))

        return dict.__getattribute__(self, attr)


def read_file(filename):
    """Try to read a given file.

    Return empty string if an error is encountered.
    """
    try:
        with codecs.open(filename, "rb", encoding="utf-8", errors="replace") as file:
            return file.read()
    except ValueError as err:
        logging.info(err)
    except Exception as err:
        logging.error(err)
    return ""


def write_file(filename, content):
    assert os.path.isabs(filename)
    try:
        with codecs.open(filename, "wb", errors="replace", encoding="utf-8") as file:
            file.write(content)
    except OSError as e:
        logging.error(f'Error while writing to "{filename}": {e}')


def make_directory(dir):
    if not os.path.isdir(dir):
        os.makedirs(dir)


def make_directories(dirs):
    for dir in dirs:
        make_directory(dir)


def make_file(file, content=""):
    if not os.path.isfile(file):
        write_file(file, content)


def make_files(file_content_pairs):
    for file, content in file_content_pairs:
        make_file(file, content)


def make_file_with_dir(file, content):
    dir = os.path.dirname(file)
    make_directory(dir)
    make_file(file, content)


def get_relative_path(from_dir, to_dir):
    """
    Try getting the relative path from from_dir to to_dir
    """
    # If the data is saved on two different windows partitions,
    # return absolute path to to_dir.
    # drive1 and drive2 are always empty strings on Unix.
    drive1, _ = os.path.splitdrive(from_dir)
    drive2, _ = os.path.splitdrive(to_dir)
    if drive1.upper() != drive2.upper():
        return to_dir

    return os.path.relpath(to_dir, from_dir)


def get_journal_title(dir):
    """
    returns the last dirname in path
    """
    dir = os.path.abspath(dir)
    # Remove double slashes and last slash
    dir = os.path.normpath(dir)

    dirname, basename = os.path.split(dir)
    # Return "/" if journal is located at /
    return basename or dirname


def get_platform_info():
    from gi.repository import GObject, Gtk
    import yaml

    functions = [
        platform.machine,
        platform.platform,
        platform.processor,
        platform.python_version,
        platform.release,
        platform.system,
    ]
    names_values = [(func.__name__, func()) for func in functions]

    names_values.extend(
        [
            (
                "GTK",
                (
                    Gtk.get_major_version(),
                    Gtk.get_minor_version(),
                    Gtk.get_micro_version(),
                ),
            ),
            ("Glib", GObject.glib_version),
            ("PyGObject", GObject.pygobject_version),
            ("YAML", yaml.__version__),
        ]
    )
    if WebKit2:
        names_values.append(
            (
                "WebKit2",
                (
                    WebKit2.get_major_version(),
                    WebKit2.get_minor_version(),
                    WebKit2.get_micro_version(),
                ),
            )
        )

    vals = [f"{name}: {val}" for name, val in names_values]
    return "System info: " + ", ".join(vals)


def system_call(args):
    """
    Asynchronous system call

    subprocess.call runs synchronously
    """
    subprocess.Popen(args)


def get_peak_memory_in_kb():
    try:
        # This will only work on Linux systems.
        with open("/proc/self/status") as status_file:
            for line in status_file:
                parts = line.split()
                if parts[0] == "VmPeak:":
                    return int(parts[1])
    except OSError:
        pass
    raise Warning("warning: could not determine peak memory")



# -----------------------------------------------------------------------
# Copyright (c) 2008-2024 Jendrik Seipp
#
# RedNotebook is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# RedNotebook is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with RedNotebook; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
# -----------------------------------------------------------------------
import os
import webbrowser

from gi.repository import GdkPixbuf, Gtk

from rednotebook import info, storage
from rednotebook.gui import format_menu, insert_menu
from rednotebook.help import help_text
from rednotebook.util import filesystem, utils


MENUBAR_XML = """\
<ui>
<menubar name="MainMenuBar">
    <menu action="Journal">
        <menuitem action="New"/>
        <menuitem action="Open"/>
        <separator/>
        <menuitem action="Save"/>
        <menuitem action="SaveAs"/>
        <separator/>
        <menuitem action="Export"/>
        <menuitem action="Backup"/>
        <menuitem action="Statistics"/>
        <separator/>
        <menuitem action="Quit"/>
    </menu>
    <menu action="Edit">
        <menuitem action="Undo"/>
        <menuitem action="Redo"/>
        <separator/>
        <menuitem action="Cut"/>
        <menuitem action="Copy"/>
        <menuitem action="Paste"/>
        <separator/>
        <menuitem action="Fullscreen"/>
        <separator/>
        <menuitem action="Find"/>
        <separator/>
        <menuitem action="CheckSpelling"/>
        <menuitem action="Options"/>
    </menu>
    {}
    {}
    <menu action="HelpMenu">
        <menuitem action="Help"/>
        <separator/>
        <menuitem action="GiveFeedback"/>
        <menuitem action="Donate"/>
        <menuitem action="Translate"/>
        <menuitem action="ReportBug"/>
        <separator/>
        <menuitem action="Info"/>
    </menu>
</menubar>
</ui>""".format(
    insert_menu.MENUBAR_XML,
    format_menu.MENUBAR_XML,
)


class MainMenuBar:
    def __init__(self, main_window, *args, **kwargs):
        self.main_window = main_window
        self.uimanager = main_window.uimanager
        self.journal = self.main_window.journal
        self.setup()

    def get_actiongroup(self):
        # Create an ActionGroup
        actiongroup = Gtk.ActionGroup("MainMenuActionGroup")

        # Create actions
        actiongroup.add_actions(
            [
                ("Journal", None, _("_Journal")),
                (
                    "New",
                    None,
                    _("New"),
                    "<Ctrl>n",
                    _("Create a new journal. The old one will be saved"),
                    self.on_new_journal_button_activate,
                ),
                (
                    "Open",
                    None,
                    _("Open"),
                    "<Ctrl>o",
                    _("Load an existing journal. The old journal will be saved"),
                    self.on_open_journal_button_activate,
                ),
                ("Save", None, _("Save"), "<Ctrl>s", None, self.on_save_button_clicked),
                (
                    "SaveAs",
                    None,
                    _("Save As"),
                    "<Ctrl><Shift>s",
                    _(
                        "Save journal at a new location. The old journal files will also be saved"
                    ),
                    self.on_save_as_menu_item_activate,
                ),
                # Translators: Verb
                (
                    "Export",
                    None,
                    _("Export"),
                    "<Ctrl>e",
                    _("Open the export assistant"),
                    self.on_export_menu_item_activate,
                ),
                # Translators: Verb
                (
                    "Backup",
                    None,
                    _("_Backup"),
                    None,
                    _("Save all the data in a zip archive"),
                    self.on_backup_activate,
                ),
                (
                    "Statistics",
                    None,
                    _("S_tatistics"),
                    None,
                    _("Show some statistics about the journal"),
                    self.on_statistics_menu_item_activate,
                ),
                (
                    "Quit",
                    None,
                    _("Quit"),
                    "<Ctrl>q",
                    _("Shutdown RedNotebook. It will not be sent to the tray."),
                    self.main_window.on_quit_activate,
                ),
                ("Edit", None, _("_Edit")),
                (
                    "Undo",
                    None,
                    _("Undo"),
                    "<Ctrl>z",
                    _("Undo text or tag edits"),
                    self.on_undo,
                ),
                (
                    "Redo",
                    None,
                    _("Redo"),
                    "<Ctrl>y",
                    _("Redo text or tag edits"),
                    self.on_redo,
                ),
                (
                    "Cut",
                    None,
                    _("Cut"),
                    "<Ctrl>x",
                    None,
                    self.on_cut_menu_item_activate,
                ),
                (
                    "Copy",
                    None,
                    _("Copy"),
                    "<Ctrl>c",
                    None,
                    self.on_copy_menu_item_activate,
                ),
                (
                    "Paste",
                    None,
                    _("Paste"),
                    "<Ctrl>v",
                    None,
                    self.on_paste_menu_item_activate,
                ),
                (
                    "Fullscreen",
                    None,
                    _("Fullscreen"),
                    "F11",
                    None,
                    self.on_fullscreen_menuitem_activate,
                ),
                (
                    "Find",
                    None,
                    _("Find"),
                    "<Ctrl>f",
                    None,
                    self.on_find_menuitem_activate,
                ),
            ]
        )
        actiongroup.add_toggle_actions(
            [
                (
                    "CheckSpelling",
                    None,
                    _("Spell Check"),
                    "F7",
                    _("Underline misspelled words"),
                    self.on_checkspelling_menuitem_toggled,
                )
            ]
        )
        actiongroup.add_actions(
            [
                (
                    "Options",
                    None,
                    _("Preferences"),
                    "<Ctrl><Alt>p",
                    None,
                    self.on_options_menuitem_activate,
                ),
                ("HelpMenu", None, _("_Help")),
                (
                    "Help",
                    None,
                    _("Contents"),
                    "<Ctrl>h",
                    _("Open the RedNotebook documentation"),
                    self.on_help_menu_item_activate,
                ),
                (
                    "Donate",
                    None,
                    _("Donate"),
                    None,
                    _("Support RedNotebook with a donation"),
                    self.on_donate,
                ),
                (
                    "Translate",
                    None,
                    _("Translate RedNotebook"),
                    None,
                    _("Help translate RedNotebook to your language"),
                    self.on_translate,
                ),
                (
                    "ReportBug",
                    None,
                    _("Report a Problem"),
                    None,
                    _("Fill out a short form about the problem"),
                    self.on_report_bug,
                ),
                (
                    "GiveFeedback",
                    None,
                    _("Give Feedback"),
                    None,
                    _("How can we improve RedNotebook?"),
                    self.on_give_feedback,
                ),
                ("Info", None, _("About"), None, None, self.on_info_activate),
            ]
        )
        return actiongroup

    def setup(self):
        # Add the actiongroup to the uimanager
        self.uimanager.insert_action_group(self.get_actiongroup(), 0)

        # Add a UI description
        self.uimanager.add_ui_from_string(MENUBAR_XML)

    def get_menu_bar(self):
        return self.uimanager.get_widget("/MainMenuBar")

    def check_journal_dir(self, action, new_dir):
        if not new_dir:
            return False
        title = _("Wrong directory")
        # Check if dir exists and is no forbidden path (e.g. $HOME dir).
        if not self.journal.dirs.is_valid_journal_path(new_dir):
            self.journal.show_message(
                _("You cannot use this directory for your journal:") + " %s" % new_dir,
                title=title,
                error=True,
            )
            return False
        if action in ["new", "saveas"] and os.listdir(new_dir):
            self.journal.show_message(
                _("Please select an empty directory."), title=title, error=True
            )
            return False
        elif action in ["open"] and not list(storage.get_journal_files(new_dir)):
            self.journal.show_message(
                _("This directory contains no journal files:") + " " + new_dir,
                title=title,
                error=True,
            )
            return False
        return True

    def select_journal(self, action, title, message):
        new_dir = self.main_window.get_new_journal_dir(title, message)
        if not self.check_journal_dir(action, new_dir):
            return

        if action == "saveas":
            self.journal.dirs.data_dir = new_dir
            self.journal.save_to_disk(saveas=True)
        self.journal.open_journal(new_dir)

    def on_new_journal_button_activate(self, widget):
        msg = "{}\n{}".format(
            _("Journals are saved in a directory, not in a single file."),
            _("The directory name will be the title of the new journal."),
        )
        self.select_journal(
            "new", _("Select an empty folder for your new journal"), msg
        )

    def on_open_journal_button_activate(self, widget):
        self.select_journal(
            "open",
            _("Select an existing journal directory"),
            _("The directory should contain your journal's data files"),
        )

    def on_save_button_clicked(self, widget):
        self.journal.save_to_disk()

    def on_save_as_menu_item_activate(self, widget):
        # widget is None when we call this method after the journal could not be
        # saved. Then another try won't succeed either.
        if widget is not None:
            self.journal.save_to_disk()
        self.select_journal(
            "saveas",
            _("Select an empty folder for the new location of your journal"),
            _("The directory name will be the new title of the journal"),
        )

    def on_undo(self, widget):
        editor = self.main_window.day_text_field
        editor.day_text_buffer.undo()

    def on_redo(self, widget):
        editor = self.main_window.day_text_field
        editor.day_text_buffer.redo()

    def _forward_signal_to_focused_widget(self, signal_name):
        widget = self.main_window.main_frame.get_focus()
        try:
            widget.emit(signal_name)
        except AttributeError:
            # No widget in focus.
            pass
        except TypeError:
            # Focus is on a widget that doesn't support this signal.
            pass

    def on_copy_menu_item_activate(self, widget):
        if self.main_window.preview_mode:
            self.main_window.html_editor.copy_to_clipboard()
        else:
            self._forward_signal_to_focused_widget("copy_clipboard")

    def on_paste_menu_item_activate(self, widget):
        self._forward_signal_to_focused_widget("paste_clipboard")

    def on_cut_menu_item_activate(self, widget):
        self._forward_signal_to_focused_widget("cut_clipboard")

    def on_fullscreen_menuitem_activate(self, widget):
        self.main_window.toggle_fullscreen()

    def on_find_menuitem_activate(self, widget):
        """
        Change to search page and put the cursor into the search box
        """
        self.main_window.search_box.entry.grab_focus()

    def on_checkspelling_menuitem_toggled(self, widget):
        self.main_window.day_text_field.enable_spell_check(widget.get_active())
        enabled = self.main_window.day_text_field.is_spell_check_enabled()
        widget.set_active(enabled)
        self.journal.config["spellcheck"] = int(enabled)

    def on_options_menuitem_activate(self, widget):
        self.main_window.options_manager.on_options_dialog()

    def on_backup_activate(self, widget):
        self.journal.archiver.backup()

    def on_export_menu_item_activate(self, widget):
        self.journal.save_old_day()
        self.main_window.export_assistant.run()

    def on_statistics_menu_item_activate(self, widget):
        self.journal.stats.show_dialog(self.main_window.stats_dialog)

    def on_help_menu_item_activate(self, widget):
        temp_dir = self.journal.dirs.temp_dir
        filesystem.write_file(os.path.join(temp_dir, "source.txt"), help_text)
        html = self.journal.convert(
            help_text,
            "html",
            headers=[_("RedNotebook Documentation"), info.version, ""],
            options={"toc": 1},
        )
        utils.show_html_in_browser(html, os.path.join(temp_dir, "help.html"))

    def on_donate(self, widget):
        webbrowser.open(info.donation_url)

    def on_translate(self, widget):
        webbrowser.open(info.translation_url)

    def on_report_bug(self, widget):
        webbrowser.open(info.bug_url)

    def on_give_feedback(self, widget):
        webbrowser.open(info.discussion_url)

    def on_info_activate(self, widget):
        self.info_dialog = self.main_window.builder.get_object("about_dialog")
        self.info_dialog.set_transient_for(self.main_window.main_frame)
        self.info_dialog.set_program_name(info.program_name)
        self.info_dialog.set_version(info.version)
        self.info_dialog.set_copyright(info.copyright_)
        self.info_dialog.set_comments(info.tagline)
        self.info_dialog.set_website(info.url)
        self.info_dialog.set_website_label(info.url)
        self.info_dialog.set_artists(info.artists)
        self.info_dialog.set_authors(info.developers)
        self.info_dialog.add_credit_section(_("Contributors:"), [info.contributors_url])
        self.info_dialog.set_translator_credits(_("translator-credits"))
        img_path = os.path.join(filesystem.image_dir, "rednotebook-icon", "rn-128.png")
        self.info_dialog.set_logo(GdkPixbuf.Pixbuf.new_from_file(img_path))
        self.info_dialog.set_license_type(Gtk.License.GPL_2_0)
        self.info_dialog.run()
        self.info_dialog.hide()



# -----------------------------------------------------------------------
# Copyright (c) 2008-2024 Jendrik Seipp
#
# RedNotebook is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# RedNotebook is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with RedNotebook; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
# -----------------------------------------------------------------------

import datetime
import logging
import os
import webbrowser

from gi.repository import GObject, Gtk


class ActionButton(Gtk.Button):
    def __init__(self, text, action):
        Gtk.Button.__init__(self, text)
        self.connect("clicked", action)


class UrlButton(ActionButton):
    def __init__(self, text, url):
        ActionButton.__init__(self, text, lambda _: webbrowser.open(url))


class CustomComboBoxEntry:
    def __init__(self, combo_box):
        self.combo_box = combo_box

        self.liststore = Gtk.ListStore(GObject.TYPE_STRING)
        self.entries = set()
        self.combo_box.set_model(self.liststore)
        self.combo_box.set_entry_text_column(0)
        self.entry = self.combo_box.get_child()

        self.entry.set_placeholder_text(_("Search"))

        # Autocompletion
        entry_completion = Gtk.EntryCompletion()
        entry_completion.set_model(self.liststore)
        entry_completion.set_minimum_key_length(1)
        entry_completion.set_text_column(0)
        self.entry.set_completion(entry_completion)

    def add_entry(self, entry):
        if entry not in self.entries:
            self.liststore.append([entry])
            self.entries.add(entry)

    def set_entries(self, value_list):
        self.clear()
        for entry in value_list:
            self.add_entry(entry)
        self.combo_box.set_model(self.liststore)

    def get_active_text(self):
        return self.entry.get_text()

    def set_active_text(self, text):
        return self.entry.set_text(text)

    def clear(self):
        self.combo_box.set_model(None)
        self.liststore.clear()
        self.entries.clear()
        self.set_active_text("")
        self.combo_box.set_model(self.liststore)


class CustomListView(Gtk.TreeView):
    def __init__(self, columns):
        """
        *columns* must be a list of (header, type) pairs e.g. [('title', str)].
        """
        Gtk.TreeView.__init__(self)
        headers, types = list(zip(*columns))
        # create a TreeStore with columns to use as the model
        self.set_model(Gtk.ListStore(*types))

        columns = [Gtk.TreeViewColumn(header) for header in headers]

        # add tvcolumns to tree_view
        for index, column in enumerate(columns):
            self.append_column(column)

            # create a CellRendererText to render the data
            cell_renderer = Gtk.CellRendererText()

            # add the cell to the tvcolumn and allow it to expand
            column.pack_start(cell_renderer, True)

            # Get markup for column, not text
            column.set_attributes(cell_renderer, markup=index)

            # Allow sorting on the column
            column.set_sort_column_id(index)

        # make it searchable
        self.set_search_column(1)


class Calendar(Gtk.Calendar):
    def __init__(self, week_numbers=False):
        Gtk.Calendar.__init__(self)
        self.set_property("show-week-numbers", week_numbers)

    def set_date(self, date):
        # Set the day temporarily to a day that is present in all months.
        self.select_day(1)

        # Gtk.Calendar show months in range [0,11].
        self.select_month(date.month - 1, date.year)

        # Select the day after the month and year have been set
        self.select_day(date.day)

    def get_date(self):
        year, month, day = Gtk.Calendar.get_date(self)
        return datetime.date(year, month + 1, day)


class Info(Gtk.InfoBar):
    icons = {
        Gtk.MessageType.ERROR: "dialog-error",
    }

    def __init__(self):
        Gtk.InfoBar.__init__(self)
        self.title_label = Gtk.Label()
        self.msg_label = Gtk.Label()
        self.title_label.set_alignment(0.0, 0.5)
        self.msg_label.set_alignment(0.0, 0.5)

        vbox = Gtk.VBox(spacing=5)
        vbox.pack_start(self.title_label, False, False, 0)
        vbox.pack_start(self.msg_label, False, False, 0)

        self.image = Gtk.Image()

        content = self.get_content_area()
        content.pack_start(self.image, False, False, 0)
        content.pack_start(vbox, False, False, 0)

        self.add_button("_Close", Gtk.ResponseType.CLOSE)
        self.connect("close", lambda x: self.hide())
        self.connect("response", self.on_response)

    def on_response(self, infobar, response_id):
        if response_id == Gtk.ResponseType.CLOSE:
            self.hide()

    def show_message(self, title, msg, msg_type):
        if not title:
            title = msg
            msg = ""
        self.title_label.set_markup(f"<b>{title}</b>")
        self.msg_label.set_markup(msg)
        self.set_message_type(msg_type)
        self.image.set_from_icon_name(
            self.icons.get(msg_type, "dialog-information"), Gtk.IconSize.DIALOG
        )
        self.show_all()


# ------------------------- Assistant Pages ------------------------------------


class AssistantPage(Gtk.VBox):
    def __init__(self, *args, **kwargs):
        GObject.GObject.__init__(self, *args, **kwargs)

        self.set_spacing(5)
        self.set_border_width(10)

        self.header = None
        self.show_all()

    def _add_header(self):
        self.header = Gtk.Label()
        self.header.set_markup("Unset")
        self.header.set_alignment(0.0, 0.5)
        self.pack_start(self.header, False, False, 0)
        self.separator = Gtk.HSeparator()
        self.pack_start(self.separator, False, False, 0)
        self.reorder_child(self.header, 0)
        self.reorder_child(self.separator, 1)
        self.show_all()

    def set_header(self, text):
        if not self.header:
            self._add_header()
        self.header.set_markup(text)


class RadioButtonPage(AssistantPage):
    def __init__(self, *args, **kwargs):
        AssistantPage.__init__(self, *args, **kwargs)

        self.buttons = []

    def add_radio_option(self, object, label, tooltip=""):
        sensitive = object.is_available()

        group = self.buttons[0] if self.buttons else None
        button = Gtk.RadioButton(group=group)
        button.set_tooltip_markup(tooltip)
        button.set_label(label)
        button.object = object
        button.set_sensitive(sensitive)
        self.pack_start(button, False, False, 0)
        self.buttons.append(button)

        if tooltip:
            description = Gtk.Label()
            description.set_alignment(0.0, 0.5)
            description.set_markup(" " * 10 + tooltip)
            description.set_sensitive(sensitive)
            self.pack_start(description, False, False, 0)

    def get_selected_object(self):
        for button in self.buttons:
            if button.get_active():
                return button.object


class PathChooserPage(AssistantPage):
    def __init__(self, assistant, *args, **kwargs):
        AssistantPage.__init__(self, *args, **kwargs)

        self.assistant = assistant

        self.last_path = None

        self.chooser = Gtk.FileChooserWidget()
        self.chooser.connect("selection-changed", self.on_path_changed)

        self.pack_start(self.chooser, True, True, 0)

    def _remove_filters(self):
        for filter in self.chooser.list_filters():
            self.chooser.remove_filter(filter)

    def prepare(self, porter):
        self._remove_filters()

        self.path_type = porter.PATHTYPE.upper()
        path = porter.DEFAULTPATH
        extension = porter.EXTENSION
        if helptext := porter.PATHTEXT:
            self.set_header(helptext)

        if self.path_type == "DIR":
            self.chooser.set_action(Gtk.FileChooserAction.SELECT_FOLDER)
        elif self.path_type == "FILE":
            self.chooser.set_action(Gtk.FileChooserAction.OPEN)
        elif self.path_type == "NEWFILE":
            self.chooser.set_action(Gtk.FileChooserAction.SAVE)
        else:
            logging.error(f'Wrong path_type "{self.path_type}"')

        if self.path_type in ["FILE", "NEWFILE"] and extension:
            filter = Gtk.FileFilter()
            filter.set_name(extension)
            filter.add_pattern(f"*.{extension}")
            self.chooser.add_filter(filter)

        if self.last_path and os.path.exists(self.last_path):
            path = self.last_path

        if os.path.isdir(path):
            self.chooser.set_current_folder(path)
        else:
            dirname, basename = os.path.split(path)
            filename, _ = os.path.splitext(basename)
            self.chooser.set_current_folder(dirname)
            self.chooser.set_current_name(f"{filename}.{extension}")

    def get_selected_path(self):
        self.last_path = self.chooser.get_filename()
        return self.last_path

    def on_path_changed(self, widget):
        return


class Assistant(Gtk.Assistant):
    def __init__(self, journal, *args, **kwargs):
        GObject.GObject.__init__(self, *args, **kwargs)

        self.journal = journal

        self.set_size_request(1000, 500)

        self.connect("cancel", self._on_cancel)
        self.connect("close", self._on_close)
        self.connect("prepare", self._on_prepare)

    def run(self):
        """
        Show assistant
        """

    def _on_cancel(self, assistant):
        """
        Cancelled -> Hide assistant
        """
        self.hide()

    def _on_close(self, assistant):
        """
        Do the action
        """

    def _on_prepare(self, assistant, page):
        """
        Called when a new page should be prepared, before it is shown
        """


class TemplateBar(Gtk.HBox):
    def __init__(self):
        GObject.GObject.__init__(self)
        self.set_spacing(2)
        label = Gtk.Label(label=f'<b>{_("Template")}</b>:')
        label.set_use_markup(True)
        self.pack_start(label, False, False, 0)
        self.save_insert_button = Gtk.Button.new_with_label(_("Save and insert"))
        self.pack_start(self.save_insert_button, False, False, 0)
        self.save_button = Gtk.Button.new_with_label(_("Save"))
        self.pack_start(self.save_button, False, False, 0)
        self.close_button = Gtk.Button.new_with_label(_("Close"))
        self.pack_start(self.close_button, False, False, 0)
        self.show_all()


class ToolbarMenuButton(Gtk.ToolButton):
    def __init__(self, icon_name, menu):
        Gtk.ToolButton.__init__(self)
        self.set_icon_name(icon_name)
        self._menu = menu
        self.connect("clicked", self._on_clicked)
        self.show_all()

    def _on_clicked(self, button):
        self._menu.popup(None, None, None, None, 0, Gtk.get_current_event_time())

    def set_menu(self, menu):
        self._menu = menu



# -----------------------------------------------------------------------
# Copyright (c) 2008-2024 Jendrik Seipp
#
# RedNotebook is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# RedNotebook is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with RedNotebook; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
# -----------------------------------------------------------------------

import logging
import os

from rednotebook.util import filesystem


def delete_comment(line):
    return "" if line.startswith("#") else line


class Config(dict):
    defaults = {
        # mingliu and MS Mincho are required for displaying Asian fonts
        # in preview mode on Windows.
        "previewFont": "Ubuntu, mingliu, MS Mincho, sans-serif",
        "autoSwitchMode": 0,
        "closeToTray": 0,
        "checkForNewVersion": 0,
        "dateTimeString": "%A, %x %X",
        "exportDateFormat": "%A, %x",
        "showTagsPane": 0,
        "weekNumbers": 0,
        "portable": 0,
        "userDir": "",
        "firstStart": 1,
        "instantSearch": 1,
        "spellcheck": 0,
        "mainFrameWidth": 1024,
        "mainFrameHeight": 700,
        "mainFrameMaximized": 0,
        "mainFrameX": None,
        "mainFrameY": None,
        "leftDividerPosition": 260,
        "rightDividerPosition": None,
        "cloudMaxTags": 1000,
        "autoIndent": 1,
    }

    obsolete_keys = {
        "useGTKMozembed",
        "useWebkit",
        "LD_LIBRARY_PATH",
        "MOZILLA_FIVE_HOME",
        "cloudTabActive",
        "mainFontSize",
        "running",
    }

    # Allow changing the value of portable only in default.cfg.
    suppressed_keys = {"portable", "user_dir"}

    def __init__(self, config_file):
        dict.__init__(self)

        self.filename = config_file

        self.update(self._read_file(self.filename))
        self.save_state()

    def save_state(self):
        """Save a copy of the dir to check for changes later"""
        self.old_config = self.copy()

    def _read_file(self, filename):
        content = filesystem.read_file(filename)

        # Delete comments and whitespace.
        lines = [delete_comment(line.strip()) for line in content.splitlines()]

        dictionary = {}

        for line in lines:
            if "=" not in line:
                continue
            pair = line.partition("=")[::2]
            key, value = (s.strip() for s in pair)
            # Skip obsolete keys to prevent rewriting them to disk.
            if key in self.obsolete_keys:
                continue

            try:
                value = int(value)
            except ValueError:
                pass

            dictionary[key] = value

        return dictionary

    def read(self, key, default=None):
        """
        Get the stored value for the given key.

        If *default* is omitted, there must be a default for *key* in
        Config.defaults. If *default* is given and there is no value
        stored for *key*, set the stored value to *default*.
        """
        if default is None:
            default = self.defaults[key]
        return self.setdefault(key, default)

    def read_list(self, key, default):
        """
        Read the string corresponding to key and convert it to a list.

        alpha,beta gamma;delta -> ['alpha', 'beta', 'gamma', 'delta']

        default should be of the form 'alpha,beta gamma;delta'
        """
        string = str(self.read(key, default))

        separators = [",", ";"]
        for separator in separators:
            string = string.replace(separator, " ")

        strings = [s.strip() for s in string.split()]
        return [s for s in strings if s]

    def write_list(self, key, list):
        self[key] = ", ".join(list)

    def changed(self):
        return self != self.old_config

    def save_to_disk(self):
        if not self.changed():
            return

        lines = [
            f"{key}={value}"
            for key, value in sorted(self.items())
            if key not in self.suppressed_keys
        ]
        try:
            filesystem.make_directory(os.path.dirname(self.filename))
            filesystem.write_file(self.filename, "\n".join(lines))
        except OSError:
            logging.error(
                "Configuration could not be saved. Please check " "your permissions"
            )
        else:
            logging.info(f"Configuration has been saved to {self.filename}")
            self.save_state()



# -----------------------------------------------------------------------
# Copyright (c) 2008-2024 Jendrik Seipp
#
# RedNotebook is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# RedNotebook is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with RedNotebook; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
# -----------------------------------------------------------------------

import ctypes
import logging
import sys

from gi.repository import Gdk, GObject, Gtk

from rednotebook.util import filesystem


_cls = None


def get_html_view_class():
    cef_disabled = True
    if cef_disabled:
        return None
    global _cls
    if not filesystem.IS_WIN:
        return None
    if not _cls:
        _cls = _make_html_view_class()
    return _cls


def _make_html_view_class():
    try:
        from cefpython3 import cefpython as cef
    except ImportError as err:
        logging.info(
            "CEF Python not found. Disabling clouds and"
            ' in-app previews. Error message: "{}"'.format(err)
        )
        return None

    class HtmlView(Gtk.DrawingArea):
        """
        Loading HTML strings only works if we pass the `url` parameter to
        CreateBrowserSync.

        When we call load_html() the first time, the browser is not yet
        created. Therefore, we store the initial html and load it when
        the browser is created.

        """

        NOTEBOOK_URL = "file:///"

        def __init__(self):
            super().__init__()
            self._browser = None
            self._win32_handle = None
            self._initial_html = ""

            sys.excepthook = cef.ExceptHook  # To shutdown CEF processes on error.
            settings = {
                "context_menu": {"enabled": False},
                # "debug": True,
                # "log_severity": cef.LOGSEVERITY_INFO,
                # "log_file": "debug.log",
            }
            cef.Initialize(settings=settings)

            GObject.threads_init()
            GObject.timeout_add(10, self.on_timer)

            self.connect("configure-event", self.on_configure)
            self.connect("size-allocate", self.on_size_allocate)
            self.connect("focus-in-event", self.on_focus_in)
            self.connect("realize", self.on_realize)

        def load_html(self, html):
            if self._browser:
                self._browser.GetMainFrame().LoadString(html, self.NOTEBOOK_URL)
            else:
                self._initial_html = html

        def set_font_size(self, size):
            pass

        def get_handle(self):
            Gdk.threads_enter()
            ctypes.pythonapi.PyCapsule_GetPointer.restype = ctypes.c_void_p
            ctypes.pythonapi.PyCapsule_GetPointer.argtypes = [ctypes.py_object]
            gpointer = ctypes.pythonapi.PyCapsule_GetPointer(
                self.get_property("window").__gpointer__, None
            )
            libgdk = ctypes.CDLL("gdk-3-vs17.dll")
            libgdk.gdk_win32_window_get_handle.argtypes = [ctypes.c_void_p]
            handle = libgdk.gdk_win32_window_get_handle(gpointer)
            Gdk.threads_leave()
            return handle

        def on_timer(self):
            cef.MessageLoopWork()
            return True

        def on_realize(self, *_):
            self._embed_browser()

        def _embed_browser(self):
            window_info = cef.WindowInfo()
            self._win32_handle = self.get_handle()
            window_info.SetAsChild(self._win32_handle)
            self._browser = cef.CreateBrowserSync(window_info, url=self.NOTEBOOK_URL)
            self._browser.SetClientCallback("OnBeforeBrowse", self.on_before_browse)
            self._browser.SetClientCallback("OnAddressChange", self.on_address_change)
            self.load_html(self._initial_html)
            self._initial_html = None

        @GObject.Signal(name="on-url-clicked", arg_types=(str,))
        def url_clicked_signal(self, url):
            logging.debug("Emitting on-url-clicked signal: %s", url)

        def on_before_browse(self, browser, frame, request, **_):
            url = request.GetUrl()
            # For some reason GetUrl() appends slash to the returned URL so we need to compensate for it:
            # (https://bugs.chromium.org/p/chromium/issues/detail?id=339054 might be the cause)
            if url == self.NOTEBOOK_URL + "/":
                # On first invocation the url points to dummy NOTEBOOK_URL.
                # There is no reason to emit signal for it.
                return False
            self.url_clicked_signal.emit(url)
            return True

        def on_address_change(self, browser, frame, url):
            if url == self.NOTEBOOK_URL:
                return
            self.url_clicked_signal.emit(url)

        def on_configure(self, *_):
            if self._browser:
                self._browser.NotifyMoveOrResizeStarted()
            return False

        def on_size_allocate(self, _, data):
            if self._browser:
                cef.WindowUtils().OnSize(self._win32_handle, 0, 0, 0)

        def on_focus_in(self, *_):
            if self._browser:
                self._browser.SetFocus(True)
                return True
            return False

        def shutdown(self, *_):
            if self._browser:
                self._browser.CloseBrowser(True)
                # Clear browser references that you keep anywhere in your
                # code. All references must be cleared for CEF to shutdown cleanly.
                self._browser = None
            cef.Shutdown()

    return HtmlView



# -----------------------------------------------------------------------
# Copyright (c) 2008-2024 Jendrik Seipp
#
# RedNotebook is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# RedNotebook is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with RedNotebook; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
# -----------------------------------------------------------------------

import datetime
import re


TEXT_RESULT_LENGTH = 42

HEX_COLOR = r"[0-9a-fA-F]{6}\b"
CPP_DIRECTIVES = "|".join(
    f"{keyword}\\b" for keyword in ("include", "define", "ifdef", "ifndef", "endif")
)
HASHTAG_EXCLUDES = "|".join((HEX_COLOR, CPP_DIRECTIVES))

ALPHA = r"[^\W\d_]"
ALPHA_NUMERIC = r"\w"
HASHTAG_TEXT = r"%(ALPHA_NUMERIC)s*%(ALPHA)s+%(ALPHA_NUMERIC)s*" % locals()
HASHTAG_PATTERN = (
    r"(^|[^%(ALPHA_NUMERIC)s&#])(#|\uFF03)(?!%(HASHTAG_EXCLUDES)s)"
    "(%(HASHTAG_TEXT)s)" % locals()
)

"""
Note:
The following regular expression is used to find hashtags in the text
and add them to the 'Tags' section on the left panel.

This pattern DOES NOT control the styling of hashtags in the text.
To control this behaviour refer to rednotebook/files/t2t.lang
(regexes) and rednotebook/files/rednotebook-highlight-style.xml (styles).

If you make changes to this pattern, is very likely you will have to
make changes to /rednotebook/files/t2t.lang
"""
HASHTAG = re.compile(HASHTAG_PATTERN, flags=re.IGNORECASE)


def escape_tag(tag):
    return tag.lower().replace(" ", "_")


def get_text_with_dots(text, start, end, found_text=None):
    """
    Find the outermost spaces and innermost newlines around
    (start, end) and add dots if needed.
    """
    bound1 = max(0, start - int(TEXT_RESULT_LENGTH // 2))
    bound2 = max(0, start)
    bound3 = min(end, len(text))
    bound4 = min(len(text), end + int(TEXT_RESULT_LENGTH // 2))

    start_values = [bound1]
    newline = text.rfind("\n", bound1, bound2)
    start_values.append(newline)
    if newline == -1:
        start_values.append(text.find(" ", bound1, bound2))
    start = max(start_values)

    end_values = [bound4]
    newline = text.find("\n", bound3, bound4)
    if newline != -1:
        end_values.append(newline)
    else:
        space = text.rfind(" ", bound3, bound4)
        if space != -1:
            end_values.append(space)
    end = min(end_values)

    assert bound1 <= start <= bound2
    assert bound3 <= end <= bound4, (bound3, end, bound4)

    res = ""
    if start > 0:
        res += "... "
    res += text[start:end]
    if end < len(text):
        res += " ..."

    res = res.replace("\n", " ")
    if found_text:
        # Make the searched_text bold
        res = res.replace(found_text, f"STARTBOLD{found_text}ENDBOLD")

    return res


def get_search_results_for_tags(text, tags):
    results = []
    add_text_to_results = False
    for line in text.splitlines():
        line = line.strip()
        if any(f"#{tag}" == line.lower() for tag in tags):
            add_text_to_results = True
        elif any(f"#{tag}" in line.lower() for tag in tags):
            for tag in tags:
                # TODO: combine into one regex and really ignore case.
                line = re.sub(rf"^#{tag}: ", "", line, re.IGNORECASE)
                line = re.sub(rf"^#{tag}:", "", line, re.IGNORECASE)
                line = re.sub(rf"^#{tag}", "", line, re.IGNORECASE)
            line = line.strip()
            results.append(line[:TEXT_RESULT_LENGTH])
    if add_text_to_results:
        results.append(get_text_with_dots(text, 0, TEXT_RESULT_LENGTH))
    return results


class Day:
    def __init__(self, month, day_number, day_content=None):
        day_content = day_content or {"text": ""}
        assert "text" in day_content, day_content

        self.month = month
        self.date = datetime.date(month.year_number, month.month_number, day_number)

        # Turn all entries of old "Tags" categories into tags without entries.
        # Apparently, "Tags" may map to None, so explicitly convert to dict.
        old_tags = day_content.pop("Tags", None) or {}
        for old_tag in old_tags:
            day_content[old_tag] = None
            self.month.edited = True

        self._content = day_content

    def _get_content(self):
        return self._content

    def _set_content(self, content):
        old_text = self.text
        new_text = content["text"]
        content["text"] = old_text
        self._content = content
        self.text = new_text

    content = property(_get_content, _set_content)

    def _get_text(self):
        """Return the day's text as a unicode string."""
        return self.content["text"]

    def _set_text(self, text):
        assert "text" in self.content
        self.content["text"] = text

    text = property(_get_text, _set_text)

    @property
    def has_text(self):
        return bool(self.text.strip())

    @property
    def empty(self):
        return len(self.content) == 1 and "text" in self.content and not self.has_text

    @property
    def hashtags(self):
        # The same tag can occur multiple times.
        return [hashtag.lower() for _, _hash, hashtag in HASHTAG.findall(self.text)]

    @property
    def categories(self):
        return list(self.get_category_content_pairs().keys())

    def get_entries(self, category):
        return sorted((self.content.get(category) or {}).keys())

    def get_category_content_pairs(self):
        """
        Returns a dict of (category: content_in_category_as_list) pairs.
        """
        pairs = {}
        for category, content in self.content.items():
            if category == "text":
                pass
            elif content is None:
                pairs[category] = []
            else:
                pairs[category] = list(content.keys())
        # Include hashtags
        for tag in self.hashtags:
            pairs[tag] = []
        return pairs

    def get_words(self, with_special_chars=False):
        categories_text = " ".join(
            " ".join([category] + content)
            for category, content in self.get_category_content_pairs().items()
        )

        all_text = f"{self.text} {categories_text}"
        words = all_text.split()

        if with_special_chars:
            return words

        # Strip all ASCII punctuation except for $, %, @ and '.
        words = [w.strip('.|-!"&/()=?*+~#_:;,<>^°`{}[]\\') for w in words]
        return [word for word in words if word]

    def get_number_of_words(self):
        return len(self.get_words(with_special_chars=True))

    def replace_all(self, old, new):
        """
        Replaces all occurrences of old content with new.
        Returns number of replacements.
        """
        self.content["text"], replaced_words = re.subn(
            re.escape(old), repl=new, string=self.content["text"], flags=re.IGNORECASE
        )
        return replaced_words

    def search(self, text, tags):
        """
        This method is only called for days that have all given tags.
        Search in date first, then in the text, then in the tags.
        Uses case-insensitive search.
        """
        results = []
        if not text:
            # Only add text result once for all tags.
            add_text_to_results = False
            for day_tag, entries in self.get_category_content_pairs().items():
                for tag in tags:
                    # We know that all tags are present, but we loop through
                    # day_tags nonetheless, to escape the day_tags.
                    if escape_tag(day_tag) != tag:
                        continue

                    if entries:
                        results.extend(entries)
                    else:
                        add_text_to_results = True
            if add_text_to_results:
                results.extend(get_search_results_for_tags(self.text, tags))
                # Handle right-pane tags without subentries.
                if not results:
                    results.append(get_text_with_dots(self.text, 0, TEXT_RESULT_LENGTH))
        elif text in str(self):
            # Date contains searched text.
            results.append(get_text_with_dots(self.text, 0, TEXT_RESULT_LENGTH))
        else:
            if text_result := self.search_in_text(text):
                results.append(text_result)
            results.extend(self.search_in_categories(text))
        return str(self), results

    def search_in_text(self, search_text):
        occurrence = self.text.upper().find(search_text.upper())

        # Check if search_text is in text
        if occurrence < 0:
            return None

        found_text = self.text[occurrence : occurrence + len(search_text)]
        return get_text_with_dots(
            self.text, occurrence, occurrence + len(search_text), found_text
        )

    def search_in_categories(self, text):
        results = []
        for category, content in self.get_category_content_pairs().items():
            if content:
                if text.upper() in category.upper():
                    results.extend(content)
                else:
                    results.extend(
                        entry for entry in content if text.upper() in entry.upper()
                    )
            elif text.upper() in category.upper():
                results.append(category)
        return results

    def __str__(self):
        return self.date.strftime("%Y-%m-%d")


class Month:
    def __init__(self, year_number, month_number, month_content=None, mtime=0):
        self.year_number = year_number
        self.month_number = month_number

        month_content = month_content or {}
        self.days = {}
        for day_number, day_content in month_content.items():
            self.days[day_number] = Day(self, day_number, day_content)

        self.edited = False
        self.mtime = mtime

    def get_day(self, day_number):
        if day_number not in self.days:
            self.days[day_number] = Day(self, day_number)
        return self.days[day_number]

    def __str__(self):
        lines = [f"Month {self.year_number} {self.month_number}"]
        lines.extend(
            f"{day_number}: {day.text}" for day_number, day in self.days.items()
        )
        return "\n".join(lines)

    @property
    def empty(self):
        return all(day.empty for day in self.days.values())



# -*- coding:utf-8 -*-
#
# Copyright (C) 2012, Maximilian Köhl <linuxmaxi@googlemail.com>
# Copyright (C) 2012, Carlos Jenkins <carlos@jenkins.co.cr>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
RedNotebook bundles its own copy of pygtkspellcheck, because pygtkspellcheck
versions < 4.0 don't decode UTF-8 from GTK widgets. The earliest Ubuntu version
with a fixed pygtkspellcheck (4.0.5) is 17.04 (see also
https://bugs.launchpad.net/rednotebook/+bug/1615629). This bug probably only
affects Python 2, so we may be able to remove our copy of pygtkspellcheck again.

A simple but quite powerful spellchecking library written in pure Python for Gtk
based on Enchant. It supports both GTK 3 and 4 via PyGObject with Python 3. For
automatic translation of the user interface it can use Gedit's translation files.
"""

import enchant
import gettext
import logging
import re
import sys
from collections import UserList

from gi.repository import Gio, GLib, GObject

# find any loaded gtk binding
if "gi.repository.Gtk" in sys.modules:
    Gtk = sys.modules["gi.repository.Gtk"]
else:
    import gi

    gi.require_version("Gtk", "3.0")
    from gi.repository import Gtk  # noqa: N813

_IS_GTK3 = Gtk.MAJOR_VERSION < 4

# public objects
__all__ = ["SpellChecker", "NoDictionariesFound"]

# logger
logger = logging.getLogger("")  # JS: was '__name__'


class NoDictionariesFound(Exception):
    """
    There aren't any dictionaries installed on the current system so
    spellchecking could not work in any way.
    """


# map between Gedit's translation and PyGtkSpellcheck's
_GEDIT_MAP = {
    "Languages": "Languages",
    "Ignore All": "Ignore _All",
    "Suggestions": "Suggestions",
    "(no suggestions)": "(no suggested words)",
    "Add to Dictionary": "Add w_ord",
    "Unknown": "Unknown",
}

_BATCHING_THRESHOLD_CHARS = 1500
_BATCH_SIZE_CHARS = 1000

# translation
if gettext.find("gedit"):
    _gedit = gettext.translation("gedit", fallback=True).gettext

    def _(message):
        return _gedit(_GEDIT_MAP[message]).replace("_", "")

else:
    locale_name = "py{}gtkspellcheck".format(sys.version_info.major)
    _ = gettext.translation(locale_name, fallback=True).gettext


def code_to_name(code, separator="_"):
    # JS: We don't bundle pylocales.
    # Simple fallback implementation for compatibility.
    # Escape underscores for GTK menuitems.
    return code.replace(separator, separator * 2)


class SpellChecker(GObject.Object):
    """
    Main spellchecking class, everything important happens here.

    :param view: GtkTextView the SpellChecker should be attached to.
    :param language: The language which should be used for spellchecking.
        Use a combination of two letter lower-case ISO 639 language code with a
        two letter upper-case ISO 3166 country code, for example en_US or de_DE.
    :param prefix: A prefix for some internal GtkTextMarks.
    :param collapse: Enclose suggestions in its own menu.
    :param params: Dictionary with Enchant broker parameters that should be set
      e.g. `enchant.myspell.dictionary.path`.

    .. attribute:: languages

        A list of supported languages.

        .. function:: exists(language)

            Checks if a language exists.

            :param language: language to check
    """

    FILTER_WORD = "word"
    FILTER_LINE = "line"
    FILTER_TEXT = "text"

    DEFAULT_FILTERS = {
        FILTER_WORD: [r"[0-9.,]+"],
        FILTER_LINE: [
            (r"(https?|ftp|file):((//)|(\\\\))+[\w\d:" r"#@%/;$()~_?+-=\\.&]+"),
            r"[\w\d]+@[\w\d.]+",
        ],
        FILTER_TEXT: [],
    }

    DEFAULT_EXTRA_CHARS = "'"

    class _LanguageList(UserList):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.mapping = dict(self)

        @classmethod
        def from_broker(cls, broker):
            return cls(
                sorted(
                    [
                        (language, code_to_name(language))
                        for language in broker.list_languages()
                    ],
                    key=lambda language: language[1],
                )
            )

        def exists(self, language):
            return language in self.mapping

    class _Mark:
        def __init__(self, buffer, name, start, iter_worker):
            self._buffer = buffer
            self._name = name
            self._mark = self._buffer.create_mark(self._name, start, True)
            self._iter_worker = iter_worker

        @property
        def iter(self):
            return self._buffer.get_iter_at_mark(self._mark)

        @property
        def inside_word(self):
            return self._iter_worker.inside_word(self.iter)

        @property
        def word(self):
            start = self.iter
            if not self._iter_worker.starts_word(start):
                self._iter_worker.backward_word_start(start)
            end = self.iter
            if self._iter_worker.inside_word(end):
                self._iter_worker.forward_word_end(end)
            return start, end

        def move(self, location):
            self._buffer.move_mark(self._mark, location)

    class _IterWorker:
        def __init__(self, extra_word_chars):
            self._extra_word_chars = extra_word_chars

        def is_extra_word_char(self, loc):
            # Language extra chararacters should also be processed once Enchant's
            # enchant_dict_get_extra_word_characters is exposed in PyEnchant

            char = loc.get_char()
            return char != "" and char in self._extra_word_chars

        def inside_word(self, loc):
            if loc.inside_word():
                return True
            elif self.starts_word(loc):
                return True
            elif loc.ends_word() and not self.ends_word(loc):
                return True
            else:
                return False

        def starts_word(self, loc):
            if loc.starts_word():
                if loc.is_start():
                    return True
                else:
                    tmp = loc.copy()
                    tmp.backward_char()
                    return not self.is_extra_word_char(tmp)
            else:
                return False

        def ends_word(self, loc):
            if loc.ends_word():
                if loc.is_end():
                    return True
                else:
                    tmp = loc.copy()
                    tmp.forward_char()
                    return not self.is_extra_word_char(tmp)
            else:
                return False

        def forward_word_end(self, loc):
            def move_through_extra_chars():
                moved = False
                while self.is_extra_word_char(loc):
                    if not loc.forward_char():
                        break
                    moved = True
                return moved

            tmp = loc.copy()
            tmp.backward_char()
            loc.forward_word_end()
            while move_through_extra_chars():
                if loc.is_end() or not loc.inside_word() or not loc.forward_word_end():
                    break

        def backward_word_start(self, loc):
            def move_through_extra_chars():
                tmp = loc.copy()
                tmp.backward_char()
                moved = False
                while self.is_extra_word_char(tmp):
                    moved = True
                    loc.assign(tmp)
                    if not tmp.backward_char():
                        break
                return moved

            loc.backward_word_start()
            while move_through_extra_chars():
                tmp = loc.copy()
                tmp.backward_char()
                if (
                    loc.is_start()
                    or not tmp.inside_word()
                    or not loc.backward_word_start()
                ):
                    break

        def sync_extra_chars(self, obj, value):
            self._extra_word_chars = obj.extra_chars

    def __init__(
        self, view, language="en", prefix="gtkspellchecker", collapse=True, params=None
    ):
        super().__init__()
        self._view = view
        self.collapse = collapse
        # GTK 3-only signals. GTK 4 uses actions, below.
        if _IS_GTK3:
            # JS: Connect to signals only after successful initialization.
            # Otherwise, functions might be called on uninitialized object.
            pass
        self._prefix = prefix
        self._broker = enchant.Broker()
        if params is not None:
            for param, value in params.items():
                self._broker.set_param(param, value)
        self.languages = SpellChecker._LanguageList.from_broker(self._broker)
        if self.languages.exists(language):
            self._language = language
        elif self.languages.exists("en"):
            logger.warning(
                (
                    'no installed dictionary for language "{}", '
                    "fallback to english".format(language)
                )
            )
            self._language = "en"
        else:
            if self.languages:
                self._language = self.languages[0][0]
                logger.warning(
                    (
                        'no installed dictionary for language "{}" '
                        "and english, fallback to first language in"
                        'language list ("{}")'
                    ).format(language, self._language)
                )
            else:
                logger.critical("no dictionaries found")
                raise NoDictionariesFound()
        self._dictionary = self._broker.request_dict(self._language)
        self._deferred_check = False
        self._filters = dict(SpellChecker.DEFAULT_FILTERS)
        self._regexes = {
            SpellChecker.FILTER_WORD: re.compile(
                "|".join(self._filters[SpellChecker.FILTER_WORD])
            ),
            SpellChecker.FILTER_LINE: re.compile(
                "|".join(self._filters[SpellChecker.FILTER_LINE])
            ),
            SpellChecker.FILTER_TEXT: re.compile(
                "|".join(self._filters[SpellChecker.FILTER_TEXT]), re.MULTILINE
            ),
        }

        self._extra_chars = SpellChecker.DEFAULT_EXTRA_CHARS
        self._iter_worker = SpellChecker._IterWorker(self._extra_chars)
        self.connect("notify::extra-chars", self._iter_worker.sync_extra_chars)

        self._batched_rechecking = False

        self._languages_menu = None
        # GTK 4-only extra menu population, gesture creation and action setup. GTK 3
        # uses signals, above.
        if not _IS_GTK3:
            extra_menu = self._view.get_extra_menu()
            if extra_menu is None:
                extra_menu = Gio.Menu()
                self._view.set_extra_menu(extra_menu)
            self._spelling_menu = Gio.Menu()
            extra_menu.append_section(None, self._spelling_menu)

            controller = Gtk.GestureClick()
            controller.set_button(0)
            controller.connect("pressed", self._gtk4_on_textview_click)
            self._view.add_controller(controller)

            self._gtk4_setup_actions()

        # JS: Connect to signals only after successful initialization.
        # Otherwise, functions might be called on uninitialized object.
        if _IS_GTK3:
            self._view.connect(
                "populate-popup", lambda entry, menu: self.populate_menu(menu)
            )
            self._view.connect("popup-menu", self._click_move_popup)
            self._view.connect("button-press-event", self._click_move_button)

        self._enabled = True
        self.buffer_initialize()

    @GObject.Property(type=str, default="")
    def language(self):
        """
        The language used for spellchecking.
        """
        return self._language

    @language.setter
    def language(self, language):
        if language != self._language and self.languages.exists(language):
            self._language = language
            self._dictionary = self._broker.request_dict(language)
            self.recheck()

    @GObject.Property(type=bool, default=False)
    def enabled(self):
        """
        Enable or disable spellchecking.
        """
        return self._enabled

    @enabled.setter
    def enabled(self, enabled):
        if enabled and not self._enabled:
            self.enable()
        elif not enabled and self._enabled:
            self.disable()

    @GObject.Property(type=bool, default=False)
    def batched_rechecking(self):
        """
        Whether to enable batched rechecking of large buffers.
        """
        return self._batched_rechecking

    @batched_rechecking.setter
    def batched_rechecking(self, val):
        self._batched_rechecking = val

    @GObject.Property(type=str, default=",")
    def extra_chars(self):
        """
        Fetch the list of extra characters beyond which words are extended.
        """
        return self._extra_chars

    @extra_chars.setter
    def extra_chars(self, chars):
        """
        Set the list of extra characters beyond which words are extended.

        :param val: String containing list of characters
        """
        self._extra_chars = chars

    def buffer_initialize(self):
        """
        Initialize the GtkTextBuffer associated with the GtkTextView. If you
        have associated a new GtkTextBuffer with the GtkTextView call this
        method.
        """
        self._buffer = self._view.get_buffer()
        self._buffer.connect("insert-text", self._before_text_insert)
        self._buffer.connect_after("insert-text", self._after_text_insert)
        self._buffer.connect_after("delete-range", self._range_delete)
        self._buffer.connect_after("mark-set", self._mark_set)
        start = self._buffer.get_bounds()[0]
        self._marks = {
            "insert-start": SpellChecker._Mark(
                self._buffer,
                "{}-insert-start".format(self._prefix),
                start,
                self._iter_worker,
            ),
            "insert-end": SpellChecker._Mark(
                self._buffer,
                "{}-insert-end".format(self._prefix),
                start,
                self._iter_worker,
            ),
            "click": SpellChecker._Mark(
                self._buffer, "{}-click".format(self._prefix), start, self._iter_worker
            ),
        }
        self._table = self._buffer.get_tag_table()

        # JS: Don't add "misspelled" tag if it's already present.
        misspelled_tag_name = "{}-misspelled".format(self._prefix)
        self._misspelled = self._table.lookup(misspelled_tag_name)
        if not self._misspelled:
            self._misspelled = Gtk.TextTag.new(misspelled_tag_name)
            self._misspelled.set_property("underline", 4)
            self._table.add(self._misspelled)

        self.ignored_tags = []

        def tag_added(tag, *args):
            if hasattr(tag, "spell_check") and not tag.spell_check:
                self.ignored_tags.append(tag)

        def tag_removed(tag, *args):
            if tag in self.ignored_tags:
                self.ignored_tags.remove(tag)

        self._table.connect("tag-added", tag_added)
        self._table.connect("tag-removed", tag_removed)
        self._table.foreach(tag_added, None)
        self.no_spell_check = self._table.lookup("no-spell-check")
        if not self.no_spell_check:
            self.no_spell_check = Gtk.TextTag.new("no-spell-check")
            self._table.add(self.no_spell_check)
        self.recheck()

    def recheck(self):
        """
        Rechecks the spelling of the whole text.
        """
        start, end = self._buffer.get_bounds()

        if self._batched_rechecking and end.get_offset() > _BATCHING_THRESHOLD_CHARS:
            start_mark = self._buffer.create_mark(None, start)
            self._continue_batched_recheck(start_mark)
        else:
            self.check_range(start, end, True)

    def disable(self):
        """
        Disable spellchecking.
        """
        self._enabled = False
        start, end = self._buffer.get_bounds()
        self._buffer.remove_tag(self._misspelled, start, end)

    def enable(self):
        """
        Enable spellchecking.
        """
        self._enabled = True
        self.recheck()

    def append_filter(self, regex, filter_type):
        """
        Append a new filter to the filter list. Filters are useful to ignore
        some misspelled words based on regular expressions.

        :param regex: The regex used for filtering.
        :param filter_type: The type of the filter.

        Filter Types:

        :const:`SpellChecker.FILTER_WORD`: The regex must match the whole word
            you want to filter. The word separation is done by Pango's word
            separation algorithm so, for example, urls won't work here because
            they are split in many words.

        :const:`SpellChecker.FILTER_LINE`: If the expression you want to match
            is a single line expression use this type. It should not be an open
            end expression because then the rest of the line with the text you
            want to filter will become correct.

        :const:`SpellChecker.FILTER_TEXT`: Use this if you want to filter
           multiline expressions. The regex will be compiled with the
           `re.MULTILINE` flag. Same with open end expressions apply here.
        """
        self._filters[filter_type].append(regex)
        if filter_type == SpellChecker.FILTER_TEXT:
            self._regexes[filter_type] = re.compile(
                "|".join(self._filters[filter_type]), re.MULTILINE
            )
        else:
            self._regexes[filter_type] = re.compile(
                "|".join(self._filters[filter_type])
            )

    def remove_filter(self, regex, filter_type):
        """
        Remove a filter from the filter list.

        :param regex: The regex which used for filtering.
        :param filter_type: The type of the filter.
        """
        self._filters[filter_type].remove(regex)
        if filter_type == SpellChecker.FILTER_TEXT:
            self._regexes[filter_type] = re.compile(
                "|".join(self._filters[filter_type]), re.MULTILINE
            )
        else:
            self._regexes[filter_type] = re.compile(
                "|".join(self._filters[filter_type])
            )

    def append_ignore_tag(self, tag):
        """
        Appends a tag to the list of ignored tags. A string will be automatic
        resolved into a tag object.

        :param tag: Tag object or tag name.
        """
        if isinstance(tag, str):
            tag = self._table.lookup(tag)
        self.ignored_tags.append(tag)

    def remove_ignore_tag(self, tag):
        """
        Removes a tag from the list of ignored tags. A string will be automatic
        resolved into a tag object.

        :param tag: Tag object or tag name.
        """
        if isinstance(tag, str):
            tag = self._table.lookup(tag)
        self.ignored_tags.remove(tag)

    def add_to_dictionary(self, word):
        """
        Adds a word to user's dictionary.

        :param word: The word to add.
        """
        self._dictionary.add_to_pwl(word)
        self.recheck()

    def ignore_all(self, word):
        """
        Ignores a word for the current session.

        :param word: The word to ignore.
        """
        self._dictionary.add_to_session(word)
        self.recheck()

    def check_range(self, start, end, force_all=False):
        """
        Checks a specified range between two GtkTextIters.

        :param start: Start iter - checking starts here.
        :param end: End iter - checking ends here.
        """
        logger.debug(
            "Check range called with range %d:%d to %d:%d and force all set to %s.",
            start.get_line(),
            start.get_line_offset(),
            end.get_line(),
            end.get_line_offset(),
            force_all,
        )
        if not self._enabled:
            return
        start = start.copy()
        end = end.copy()
        if self._iter_worker.inside_word(end):
            self._iter_worker.forward_word_end(end)
        if self._iter_worker.inside_word(start) or self._iter_worker.ends_word(start):
            self._iter_worker.backward_word_start(start)
        if not self._iter_worker.starts_word(start):
            self._iter_worker.forward_word_end(start)
            self._iter_worker.backward_word_start(start)
        self._buffer.remove_tag(self._misspelled, start, end)
        cursor = self._buffer.get_iter_at_mark(self._buffer.get_insert())
        precursor = cursor.copy()
        precursor.backward_char()
        highlight = cursor.has_tag(self._misspelled) or precursor.has_tag(
            self._misspelled
        )
        word_start = start.copy()
        while word_start.compare(end) < 0:
            word_end = word_start.copy()
            self._iter_worker.forward_word_end(word_end)
            in_word = (word_start.compare(cursor) < 0) and (
                cursor.compare(word_end) <= 0
            )
            if in_word and not force_all:
                if highlight:
                    self._check_word(word_start, word_end)
                else:
                    self._deferred_check = True
            else:
                self._check_word(word_start, word_end)
                self._deferred_check = False
            self._iter_worker.forward_word_end(word_end)
            self._iter_worker.backward_word_start(word_end)
            if word_start.equal(word_end):
                break
            word_start = word_end.copy()

    def populate_menu(self, menu):
        """
        Populate the provided menu with spelling items.

        :param menu: The menu to populate.
        """
        # In GTK 4 our existing menu needs to be cleared, providing for disabling
        if not _IS_GTK3:
            menu.remove_all()

        if not self._enabled:
            return

        if _IS_GTK3:
            separator = Gtk.SeparatorMenuItem.new()
            separator.show()
            menu.prepend(separator)
            languages = Gtk.MenuItem.new_with_label(_("Languages"))
            languages.set_submenu(self._get_languages_menu())
            languages.show_all()
            menu.prepend(languages)
        else:
            menu.append_item(self._get_languages_menu())

        if self._marks["click"].inside_word:
            start, end = self._marks["click"].word
            if start.has_tag(self._misspelled):
                word = self._buffer.get_text(start, end, False)
                items = self._suggestion_menu(word)
                if self.collapse:
                    menu_label = _("Suggestions")
                    if _IS_GTK3:
                        suggestions = Gtk.MenuItem.new_with_label(menu_label)
                        submenu = Gtk.Menu.new()
                    else:
                        suggestions = Gio.MenuItem.new(menu_label, None)
                        submenu = Gio.Menu.new()
                    for item in items:
                        if _IS_GTK3:
                            submenu.append(item)
                        else:
                            submenu.append_item(item)
                    suggestions.set_submenu(submenu)
                    if _IS_GTK3:
                        suggestions.show_all()
                        menu.prepend(suggestions)
                    else:
                        menu.prepend_item(suggestions)
                else:
                    items.reverse()
                    for item in items:
                        if _IS_GTK3:
                            menu.prepend(item)
                            menu.show_all()
                        else:
                            menu.prepend_item(item)

    def move_click_mark(self, iter):
        """
        Move the "click" mark, used to determine the word being checked.

        :param iter: TextIter for the new location
        """
        self._marks["click"].move(iter)

    def _gtk4_setup_actions(self) -> None:
        action_group = Gio.SimpleActionGroup.new()

        action = Gio.SimpleAction.new("ignore-all", GLib.VariantType("s"))
        action.connect(
            "activate", lambda _action, word: self.ignore_all(word.get_string())
        )
        action_group.add_action(action)

        action = Gio.SimpleAction.new("add-to-dictionary", GLib.VariantType("s"))
        action.connect(
            "activate", lambda _action, word: self.add_to_dictionary(word.get_string())
        )
        action_group.add_action(action)

        action = Gio.SimpleAction.new("replace-word", GLib.VariantType("s"))
        action.connect(
            "activate",
            lambda _action, suggestion: self._replace_word(suggestion.get_string()),
        )
        action_group.add_action(action)

        language = Gio.PropertyAction.new("language", self, "language")
        action_group.add_action(language)

        self._view.insert_action_group("spelling", action_group)

    def _get_languages_menu(self):
        if _IS_GTK3:
            return self._build_languages_menu()
        else:
            if self._languages_menu is None:
                self._languages_menu = self._build_languages_menu()
            return self._languages_menu

    def _build_languages_menu(self):
        if _IS_GTK3:

            def _set_language(item, code):
                self.language = code

            menu = Gtk.Menu.new()
            group = []
            connect = []
        else:
            menu = Gio.Menu.new()

        for code, name in self.languages:
            if _IS_GTK3:
                item = Gtk.RadioMenuItem.new_with_label(group, name)
                group.append(item)
                if code == self.language:
                    item.set_active(True)
                connect.append((item, code))
                menu.append(item)
            else:
                item = Gio.MenuItem.new(name, None)
                item.set_action_and_target_value(
                    "spelling.language", GLib.Variant.new_string(code)
                )
                menu.append_item(item)
        if _IS_GTK3:
            for item, code in connect:
                item.connect("activate", _set_language, code)
            return menu
        else:
            return Gio.MenuItem.new_submenu(_("Languages"), menu)

    def _suggestion_menu(self, word):
        menu = []
        suggestions = self._dictionary.suggest(word)
        if not suggestions:
            # Show GTK 3 no suggestions item (removed for GTK 4)
            if _IS_GTK3:
                item = Gtk.MenuItem.new()
                label = Gtk.Label.new("")
                try:
                    label.set_halign(Gtk.Align.LEFT)
                except AttributeError:
                    label.set_alignment(0.0, 0.5)
                label.set_markup("<i>{text}</i>".format(text=_("(no suggestions)")))
                item.add(label)
                menu.append(item)
        else:
            for suggestion in suggestions:
                if _IS_GTK3:
                    item = Gtk.MenuItem.new()
                    label = Gtk.Label.new("")
                    label.set_markup("<b>{text}</b>".format(text=suggestion))
                    try:
                        label.set_halign(Gtk.Align.LEFT)
                    except AttributeError:
                        label.set_alignment(0.0, 0.5)
                    item.add(label)

                    def _make_on_activate(word):
                        return lambda *args: self._replace_word(word)

                    item.connect("activate", _make_on_activate(word))
                else:
                    escaped = suggestion.replace("'", "\\'")
                    item = Gio.MenuItem.new(
                        suggestion, f"spelling.replace-word('{escaped}')"
                    )
                menu.append(item)
        add_to_dict_menu_label = _("Add to Dictionary")
        word_escaped = word.replace("'", "\\'")
        if _IS_GTK3:
            menu.append(Gtk.SeparatorMenuItem.new())
            item = Gtk.MenuItem.new_with_label(add_to_dict_menu_label)
            item.connect("activate", lambda *args: self.add_to_dictionary(word))
        else:
            item = Gio.MenuItem.new(
                add_to_dict_menu_label, f"spelling.add-to-dictionary('{word_escaped}')"
            )
        menu.append(item)
        ignore_menu_label = _("Ignore All")
        if _IS_GTK3:
            item = Gtk.MenuItem.new_with_label(ignore_menu_label)
            item.connect("activate", lambda *args: self.ignore_all(word))
        else:
            item = Gio.MenuItem.new(
                ignore_menu_label, f"spelling.ignore-all('{word_escaped}')"
            )
        menu.append(item)
        return menu

    def _click_move_popup(self, *args):
        self.move_click_mark(self._buffer.get_iter_at_mark(self._buffer.get_insert()))
        return False

    def _click_move_button(self, widget, event):
        if event.button == 3:
            self._move_mark_for_input(event.x, event.y)
        return False

    def _move_mark_for_input(self, input_x, input_y):
        if self._deferred_check:
            self._check_deferred_range(True)
        x, y = self._view.window_to_buffer_coords(2, int(input_x), int(input_y))
        iter = self._view.get_iter_at_location(x, y)
        if isinstance(iter, tuple):
            iter = iter[1]
        self.move_click_mark(iter)

    def _gtk4_on_textview_click(self, click, n_press, x, y) -> None:
        if n_press != 1 or click.get_current_button() != 3:
            return

        self._move_mark_for_input(x, y)
        self.populate_menu(self._spelling_menu)

    def _before_text_insert(self, textbuffer, location, text, length):
        self._marks["insert-start"].move(location)

    def _after_text_insert(self, textbuffer, location, text, length):
        start = self._marks["insert-start"].iter
        self.check_range(start, location)
        self._marks["insert-end"].move(location)

    def _range_delete(self, textbuffer, start, end):
        self.check_range(start, end)

    def _mark_set(self, textbuffer, location, mark):
        if mark == self._buffer.get_insert() and self._deferred_check:
            self._check_deferred_range(False)

    def _replace_word(self, new_word):
        start, end = self._marks["click"].word
        old_word = start.get_text(end)
        offset = start.get_offset()
        self._buffer.begin_user_action()
        self._buffer.delete(start, end)
        self._buffer.insert(self._buffer.get_iter_at_offset(offset), new_word)
        self._buffer.end_user_action()
        self._dictionary.store_replacement(old_word, new_word)

    def _check_deferred_range(self, force_all):
        start = self._marks["insert-start"].iter
        end = self._marks["insert-end"].iter
        self.check_range(start, end, force_all)

    def _check_word(self, start, end):
        if start.has_tag(self.no_spell_check):
            return
        for tag in self.ignored_tags:
            if start.has_tag(tag):
                return
        word = self._buffer.get_text(start, end, False).strip()
        logger.debug(
            "Checking word %s in range %d:%d to %d:%d.",
            word,
            start.get_line(),
            start.get_line_offset(),
            end.get_line(),
            end.get_line_offset(),
        )
        if not word:
            return
        if len(self._filters[SpellChecker.FILTER_WORD]):
            if self._regexes[SpellChecker.FILTER_WORD].match(word):
                return
        if len(self._filters[SpellChecker.FILTER_LINE]):
            if _IS_GTK3:
                line_start = self._buffer.get_iter_at_line(start.get_line())
            else:
                _success, line_start = self._buffer.get_iter_at_line(start.get_line())
            line_end = end.copy()
            line_end.forward_to_line_end()
            line = self._buffer.get_text(line_start, line_end, False)
            for match in self._regexes[SpellChecker.FILTER_LINE].finditer(line):
                if match.start() <= start.get_line_offset() <= match.end():
                    if _IS_GTK3:
                        start = self._buffer.get_iter_at_line_offset(
                            start.get_line(), match.start()
                        )
                        end = self._buffer.get_iter_at_line_offset(
                            start.get_line(), match.end()
                        )
                    else:
                        # Success is not verified here as the locations come directly
                        # from the buffer
                        _success, start = self._buffer.get_iter_at_line_offset(
                            start.get_line(), match.start()
                        )
                        _success, end = self._buffer.get_iter_at_line_offset(
                            start.get_line(), match.end()
                        )
                    self._buffer.remove_tag(self._misspelled, start, end)
                    return
        if len(self._filters[SpellChecker.FILTER_TEXT]):
            text_start, text_end = self._buffer.get_bounds()
            text = self._buffer.get_text(text_start, text_end, False)
            for match in self._regexes[SpellChecker.FILTER_TEXT].finditer(text):
                if match.start() <= start.get_offset() <= match.end():
                    start = self._buffer.get_iter_at_offset(match.start())
                    end = self._buffer.get_iter_at_offset(match.end())
                    self._buffer.remove_tag(self._misspelled, start, end)
                    return
        if not self._dictionary.check(word):
            self._buffer.apply_tag(self._misspelled, start, end)

    def _continue_batched_recheck(self, start_mark):
        if start_mark.get_buffer() != self._buffer:
            return
        start = self._buffer.get_iter_at_mark(start_mark)
        self._buffer.delete_mark(start_mark)

        if not self._enabled:
            return

        end = start.copy()
        end.forward_chars(_BATCH_SIZE_CHARS)
        self._iter_worker.forward_word_end(end)

        self.check_range(start, end, True)

        if not end.is_end():
            end.forward_char()
            start_mark = self._buffer.create_mark(None, end)
            GLib.idle_add(self._continue_batched_recheck, start_mark)



# -----------------------------------------------------------------------
# Copyright (c) 2008-2024 Jendrik Seipp
#
# RedNotebook is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# RedNotebook is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with RedNotebook; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
# -----------------------------------------------------------------------

import argparse
import builtins


if not hasattr(builtins, "_"):

    def _(string):
        return string


program_name = "RedNotebook"
tagline = _("A Desktop Journal")
version = "2.41"
author = "Jendrik Seipp"
author_mail = "jendrikseipp@gmail.com"
copyright_ = "Copyright (c) 2008-2024 Jendrik Seipp"
url = "https://rednotebook.app"
downloads_url = "https://rednotebook.app/downloads.html"
donation_url = "https://rednotebook.app/downloads.html"
translation_url = "https://hosted.weblate.org/engage/rednotebook/"
bug_url = "https://github.com/jendrikseipp/rednotebook/issues"
version_url = "https://raw.githubusercontent.com/jendrikseipp/rednotebook/stable/rednotebook/info.py"
contributors_url = "https://github.com/jendrikseipp/rednotebook/graphs/contributors"
discussion_url = "https://github.com/jendrikseipp/rednotebook/discussions"

developers = ["%(author)s <%(author_mail)s>" % locals()]
artists = ["Ciaran"]

comments = _(
    """\
RedNotebook is a modern desktop journal. It lets you format, tag and
search your entries. You can also add pictures, links and customizable
templates, spell check your notes, and export to plain text, HTML or
Latex.
"""
)

journal_path_help = """\
(optional) Specify the directory storing the journal data.
The journal argument can be one of the following:
 - An absolute path (e.g. /home/username/myjournal)
 - A relative path (e.g. ../dir/myjournal)
 - The name of a directory under $HOME/.rednotebook/ (e.g. myjournal)

If the journal argument is omitted then the last session's journal
path will be used. At the first program start, this defaults to
"$HOME/.rednotebook/data".
"""


def get_commandline_parser():
    parser = argparse.ArgumentParser(
        description=comments, formatter_class=argparse.RawTextHelpFormatter
    )
    parser.add_argument("--version", action="version", version=f"RedNotebook {version}")
    parser.add_argument(
        "--date", dest="start_date", help="load specified date (format: YYYY-MM-DD)"
    )
    parser.add_argument("journal", nargs="?", help=journal_path_help)
    return parser


desktop_file = """\
[Desktop Entry]
Version=1.0
Name=RedNotebook
GenericName=Journal
Comment=Daily journal with calendar, templates and keyword search
Exec=rednotebook
Icon=rednotebook
Terminal=false
Type=Application
Categories=Office;
StartupNotify=true
"""



# -----------------------------------------------------------------------
# Copyright (c) 2008-2024 Jendrik Seipp
#
# RedNotebook is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# RedNotebook is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with RedNotebook; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
# -----------------------------------------------------------------------

import logging
import os
import urllib.request

from gi.repository import GObject, Gtk, Pango

from rednotebook.util import filesystem


try:
    from rednotebook.external import spellcheck
except ImportError:
    logging.warning("For spell checking, please install enchant (python3-enchant).")
    spellcheck = None


try:
    DEFAULT_FONT = Gtk.Settings.get_default().get_property("gtk-font-name")
except AttributeError:
    # Gtk.Settings.get_default() returns None on the CI systems without a screen.
    DEFAULT_FONT = "Ubuntu 10"


class Editor(GObject.GObject):
    __gsignals__ = {"can-undo-redo-changed": (GObject.SIGNAL_RUN_FIRST, None, ())}

    def __init__(self, day_text_view):
        super().__init__()
        self.day_text_view = day_text_view

        self._connect_undo_signals()

        self.search_text = ""

        # spell checker
        self._spell_checker = None
        self.enable_spell_check(False)

        # Enable drag&drop
        self.day_text_view.connect("drag-data-received", self.on_drag_data_received)

        # Sometimes making the editor window very small causes the program to freeze
        # So we forbid that behaviour, by setting a minimum width
        self.day_text_view.set_size_request(1, -1)

        self.font = Pango.FontDescription(DEFAULT_FONT)
        self.default_size = self.font.get_size() / Pango.SCALE
        logging.debug(f"Default font: {self.font.to_string()}")
        logging.debug(f"Default size: {self.default_size}")

    def replace_buffer(self, buffer):
        self.day_text_view.set_buffer(buffer)
        # Initialize buffer only if it is new.
        if self._spell_checker:
            self._spell_checker.buffer_initialize()
        self._connect_undo_signals()
        self._can_undo_redo_changed()

    @property
    def day_text_buffer(self):
        return self.day_text_view.get_buffer()

    def _connect_undo_signals(self):
        undo_mgr = self.day_text_buffer.get_undo_manager()
        undo_mgr.connect("can-undo-changed", self._can_undo_redo_changed)
        undo_mgr.connect("can-redo-changed", self._can_undo_redo_changed)

    def set_text(self, text, undoing=False):
        # We typically don't want to be able to undo/redo a replacement of the
        # whole text, so we mark it as 'not undoable'.
        self.day_text_buffer.begin_not_undoable_action()
        self.insert(text, overwrite=True, undoing=undoing)
        self.day_text_buffer.end_not_undoable_action()

    def get_text(self, iter_start=None, iter_end=None):
        iter_start = iter_start or self.day_text_buffer.get_start_iter()
        iter_end = iter_end or self.day_text_buffer.get_end_iter()
        return self.day_text_buffer.get_text(iter_start, iter_end, True)

    def insert(self, text, iter=None, overwrite=False, undoing=False):
        if overwrite:
            self.day_text_buffer.set_text("")
            iter = self.day_text_buffer.get_start_iter()

        if iter is None:
            self.day_text_buffer.insert_at_cursor(text)
        else:
            if isinstance(iter, Gtk.TextMark):
                iter = self.day_text_buffer.get_iter_at_mark(iter)
            self.day_text_buffer.insert(iter, text)

    def replace_selection(self, text):
        self.day_text_buffer.delete_selection(interactive=False, default_editable=True)
        self.day_text_buffer.insert_at_cursor(text)

    def replace_selection_and_highlight(self, p1, p2, p3):
        """
        Insert all three parts and highlight the middle part.
        """
        self.replace_selection(p1 + p2 + p3)
        # Get the mark at the end of the insertion.
        insert_mark = self.day_text_buffer.get_insert()
        insert_iter = self.day_text_buffer.get_iter_at_mark(insert_mark)
        start = insert_iter.copy()
        end = insert_iter.copy()
        start.backward_chars(len(p3) + len(p2))
        end.backward_chars(len(p3))
        self.day_text_buffer.select_range(start, end)

    def highlight(self, text):
        self.search_text = text
        buf = self.day_text_buffer

        # Clear previous highlighting
        start = buf.get_start_iter()
        end = buf.get_end_iter()
        buf.remove_tag_by_name("highlighter", start, end)

        # Highlight matches
        if text:
            for match_start, match_end in self.iter_search_matches(text):
                buf.apply_tag_by_name("highlighter", match_start, match_end)

    search_flags = (
        Gtk.TextSearchFlags.VISIBLE_ONLY | Gtk.TextSearchFlags.CASE_INSENSITIVE
    )

    def iter_search_matches(self, text):
        it = self.day_text_buffer.get_start_iter()
        while True:
            match = it.forward_search(text, self.search_flags)
            if not match:
                return
            yield match
            it = match[1]  # Continue searching from after the match

    def scroll_to_text(self, text):
        for match_start, _ in self.iter_search_matches(text):
            # It is safer to scroll to a mark than an iter
            mark = self.day_text_buffer.create_mark(
                "highlight_query", match_start, left_gravity=False
            )
            self.day_text_view.scroll_to_mark(mark, 0, False, 0, 0)
            self.day_text_buffer.delete_mark(mark)
            return  # Stop after the first match

    def get_selected_text(self):
        if bounds := self.day_text_buffer.get_selection_bounds():
            return self.get_text(*bounds)
        else:
            return ""

    def get_text_left_of_selection(self, length):
        bounds = self.get_selection_bounds()
        start = bounds[0].copy()
        start.backward_chars(length)
        end = bounds[0]
        return self.get_text(start, end)

    def get_text_right_of_selection(self, length):
        bounds = self.get_selection_bounds()
        start = bounds[1]
        end = bounds[1].copy()
        end.forward_chars(length)
        return self.get_text(start, end)

    @staticmethod
    def sort_iters(*iters):
        return sorted(iters, key=lambda iter: iter.get_offset())

    def get_selection_bounds(self):
        """
        Return sorted iters

        Do not mix this method up with the textbuffer's method of the same name
        That method returns an empty tuple, if there is no selection
        """
        mark1 = self.day_text_buffer.get_insert()
        mark2 = self.day_text_buffer.get_selection_bound()
        iter1 = self.day_text_buffer.get_iter_at_mark(mark1)
        iter2 = self.day_text_buffer.get_iter_at_mark(mark2)
        return self.sort_iters(iter1, iter2)

    def _get_markups(self, format, selection):
        format_to_markups = {
            "bold": ("**", "**"),
            "italic": ("//", "//"),
            "monospace": ("``", "``"),
            "underline": ("__", "__"),
            "strikethrough": ("--", "--"),
        }

        left_markup, right_markup = format_to_markups[format]
        if format == "monospace" and "\n" in selection:
            left_markup = "\n```\n"
            right_markup = "\n```\n"
        return left_markup, right_markup

    def apply_format(self, format):
        selection = self.get_selected_text()
        left_markup, right_markup = self._get_markups(format, self.get_selected_text())

        # Apply formatting only once.
        if self.get_text_left_of_selection(
            len(left_markup)
        ) == left_markup or selection.startswith(left_markup):
            left_markup = ""
        if self.get_text_right_of_selection(
            len(right_markup)
        ) == right_markup or selection.endswith(right_markup):
            right_markup = ""

        # Don't add unneeded newlines.
        if left_markup.startswith("\n") and self.get_text_left_of_selection(1) in [
            "\n",
            "",
        ]:
            left_markup = left_markup[1:]
        if right_markup.endswith("\n") and self.get_text_right_of_selection(1) in [
            "\n",
            "",
        ]:
            right_markup = right_markup[:-1]

        text = selection or " "
        self.replace_selection_and_highlight(left_markup, text, right_markup)
        self.day_text_view.grab_focus()

    def set_font(self, font_name):
        font = Pango.FontDescription(font_name)
        self.day_text_view.modify_font(font)

    def hide(self):
        self.day_text_view.hide()

    # ===========================================================
    # Spell checking.

    def can_spell_check(self):
        """Return True if spell checking is available."""
        return spellcheck is not None

    def is_spell_check_enabled(self):
        return bool(self._spell_checker and self._spell_checker.enabled)

    def _enable_spell_check(self):
        assert self.can_spell_check()
        if self._spell_checker:
            self._spell_checker.enable()
        else:
            try:
                self._spell_checker = spellcheck.SpellChecker(
                    self.day_text_view, filesystem.LANGUAGE
                )
            except spellcheck.NoDictionariesFound:
                logging.warning("No spell checking dictionaries found.")
                self._spell_checker = None
            except Exception as err:
                logging.error(
                    "Spell checking could not be enabled. %s: %s"
                    % (type(err).__name__, err)
                )
                self._spell_checker = None

    def _disable_spell_check(self):
        if self._spell_checker:
            self._spell_checker.disable()

    def enable_spell_check(self, enable=True):
        """Enable/disable spell check."""
        if not self.can_spell_check():
            return

        if enable:
            self._enable_spell_check()
        else:
            self._disable_spell_check()

    # ===========================================================

    def on_drag_data_received(
        self, widget, drag_context, x, y, selection, info, timestamp
    ):
        # We do not want the default behaviour
        self.day_text_view.emit_stop_by_name("drag-data-received")

        iter = self.day_text_view.get_iter_at_location(x, y)[1]

        def is_pic(uri):
            _, ext = os.path.splitext(uri)
            return ext.lower().strip(".") in "png jpeg jpg gif eps bmp svg".split()

        uris = (selection.get_text() or "").split()
        logging.debug(f"Text: {selection.get_text()}")
        logging.debug(f"URIs: {uris}")
        for uri in uris:
            uri = uri.strip()
            uri = urllib.request.url2pathname(uri)
            dirs, filename = os.path.split(uri)
            uri_without_ext, ext = os.path.splitext(uri)
            if is_pic(uri):
                self.insert(f'[""{uri_without_ext}""{ext}]\n', iter)
            else:
                # It is always safer to add the "file://" protocol and the ""s
                self.insert(f'[{filename} ""{uri}""]\n', iter)

        drag_context.finish(True, False, timestamp)
        # No further processing
        return True

    def _can_undo_redo_changed(self, undo_mgr=None):
        self.emit("can-undo-redo-changed")



# -----------------------------------------------------------------------
# Copyright (c) 2008-2024 Jendrik Seipp
#
# RedNotebook is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# RedNotebook is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with RedNotebook; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
# -----------------------------------------------------------------------

from gi.repository import Gtk

from rednotebook.gui import customwidgets


MENUITEMS_XML = """\
    <menuitem action="Bold"/>
    <menuitem action="Italic"/>
    <menuitem action="Monospace"/>
    <menuitem action="Underline"/>
    <menuitem action="Strikethrough"/>
    <menuitem action="Clear"/>
"""

TOOLBAR_XML = (
    """\
<ui>
<popup action="FormatMenu">
%s
</popup>
</ui>"""
    % MENUITEMS_XML
)

MENUBAR_XML = (
    """\
<menu action="FormatMenuBar">
%s
</menu>
"""
    % MENUITEMS_XML
)


class FormatMenu:
    FORMAT_TO_MARKUP = {
        "bold": "**",
        "italic": "//",
        "monospace": "``",
        "underline": "__",
        "strikethrough": "--",
    }

    def __init__(self, main_window):
        self.main_window = main_window
        self.setup()

    def setup(self):
        uimanager = self.main_window.uimanager

        # Create an ActionGroup
        actiongroup = Gtk.ActionGroup("FormatActionGroup")

        def apply_format(action):
            format_ = action.get_name().lower()
            iter_ = self.main_window.categories_tree_view.get_selected_node()
            if iter_:
                markup = self.FORMAT_TO_MARKUP[format_]
                text = self.main_window.categories_tree_view.get_iter_value(iter_)
                text = f"{markup}{text}{markup}"
                self.main_window.categories_tree_view.set_iter_value(iter_, text)
            else:
                self.main_window.day_text_field.apply_format(format_)

        def shortcut(char):
            # Translators: The Control (Ctrl) key
            return ""  # return ' (%s+%s)' % (_('Ctrl'), char)

        # Create actions
        actions = [
            (
                "Bold",
                None,
                _("Bold") + shortcut("B"),
                "<Control>B",
                None,
                apply_format,
            ),
            (
                "Italic",
                None,
                _("Italic") + shortcut("I"),
                "<Control>I",
                None,
                apply_format,
            ),
            (
                "Monospace",
                None,
                _("Monospace") + shortcut("M"),
                "<Control>M",
                None,
                apply_format,
            ),
            (
                "Underline",
                None,
                _("Underline") + shortcut("U"),
                "<Control>U",
                None,
                apply_format,
            ),
            (
                "Strikethrough",
                None,
                _("Strikethrough") + shortcut("K"),
                "<Control>K",
                None,
                apply_format,
            ),
            (
                "Clear",
                None,
                _("Clear Format") + shortcut("R"),
                "<Control>R",
                None,
                self.on_clear_format,
            ),
            # Translators: Noun
            ("FormatMenuBar", None, _("_Format")),
        ]

        actiongroup.add_actions(actions)

        # Add the actiongroup to the uimanager
        uimanager.insert_action_group(actiongroup, 0)

        # Add a UI description
        uimanager.add_ui_from_string(TOOLBAR_XML)

        # Create a Menu
        menu = uimanager.get_widget("/FormatMenu")

        self.main_window.format_button = customwidgets.ToolbarMenuButton(
            "format-text-bold", menu
        )
        # Translators: Noun
        self.main_window.format_button.set_label(_("Format"))
        tip = _("Format the selected text or tag")
        self.main_window.format_button.set_tooltip_text(tip)
        self.main_window.builder.get_object("edit_toolbar").insert(
            self.main_window.format_button, -1
        )
        self.main_window.format_actiongroup = actiongroup

    def on_clear_format(self, action):
        editor = self.main_window.day_text_field
        sel_text = editor.get_selected_text()
        for markup in list(self.FORMAT_TO_MARKUP.values()) + ["=== ", " ==="]:
            sel_text = sel_text.replace(markup, "")
        editor.replace_selection(sel_text)



# -----------------------------------------------------------------------
# Copyright (c) 2008-2024 Jendrik Seipp
#
# RedNotebook is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# RedNotebook is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with RedNotebook; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
# -----------------------------------------------------------------------

import http.client
import locale
import logging
import os.path
import re
import signal
import threading
from urllib.request import urlopen
import webbrowser

from gi.repository import GLib, Gtk

from rednotebook import info
from rednotebook.util import filesystem


def sort_asc(string):
    return str(string).lower()


def setup_signal_handlers(journal):
    """
    Catch abnormal exits of the program and save content to disk
    Look in signal man page for signal names

    SIGKILL cannot be caught
    SIGINT is caught again by KeyboardInterrupt
    """
    signals = []
    signal_names = [
        "SIGHUP",  # Terminal closed, Parent process dead
        "SIGINT",  # Interrupt from keyboard (CTRL-C)
        "SIGQUIT",  # Quit from keyboard
        "SIGABRT",  # Abort signal from abort(3)
        "SIGTERM",  # Termination signal
        "SIGTSTP",  # Stop typed at tty
    ]

    def signal_handler(signum, frame):
        logging.info(f"Program was abnormally aborted with signal {signum}")
        journal.exit()

    for signal_name in signal_names:
        signal_number = getattr(signal, signal_name, None)
        if signal_number is not None:
            try:
                signal.signal(signal_number, signal_handler)
                signals.append(signal_number)
            except RuntimeError:
                logging.info("Could not connect signal number %d" % signal_number)

    logging.info(f"Connected Signals: {signals}")


def get_gtk_colors(widget):
    """Retrieve colors of the current GTK theme for the given widget.

    The get_background_color() method is deprecated, but I couldn't find
    a different way for retrieving the color.
    """
    style = widget.get_style_context()
    bg_color = style.get_background_color(Gtk.StateFlags.NORMAL).to_string()
    fg_color = style.get_color(Gtk.StateFlags.NORMAL).to_string()
    logging.debug(f"Background color: {bg_color}")
    logging.debug(f"Foreground color: {fg_color}")
    return bg_color, fg_color


def get_new_version_number():
    """
    Read version number from website and return None if it cannot be read.
    """
    version_pattern = re.compile(r'^version = "(.+)"$', flags=re.M)

    try:
        project_xml = urlopen(info.version_url).read()
    except (OSError, http.client.HTTPException):
        return None

    project_xml = project_xml.decode("utf-8")
    match = version_pattern.search(project_xml)
    if not match:
        return None
    new_version = match[1]
    logging.info(f"Latest version: {new_version}")
    return new_version


def _show_update_dialog(journal, current_version, new_version, startup):
    dialog = Gtk.MessageDialog(
        parent=None,
        flags=Gtk.DialogFlags.MODAL,
        type=Gtk.MessageType.INFO,
        buttons=Gtk.ButtonsType.YES_NO,
        message_format=None,
    )
    dialog.set_transient_for(journal.frame.main_frame)
    primary_text = (
        _("You have version <b>%s</b>.") % current_version
        + " "
        + _("The latest version is <b>%s</b>.") % new_version
    )
    secondary_text = (
        _("If you like the program, please consider making a donation.")
        + " "
        + _("Do you want to visit the RedNotebook homepage?")
    )
    dialog.set_markup(primary_text)
    dialog.format_secondary_text(secondary_text)

    # Let user disable checks
    response_not_again = 30
    if startup:
        dialog.add_button(_("Do not ask again"), response_not_again)

    response = dialog.run()
    dialog.hide()

    if response == Gtk.ResponseType.YES:
        webbrowser.open(info.downloads_url)
    elif response == response_not_again:
        logging.info("Checks for new versions disabled")
        journal.config["checkForNewVersion"] = 0


def _get_version_tuple(version):
    parts = [int(x) for x in version.strip(" .").split(".")]
    assert len(parts) <= 3, parts
    while len(parts) < 3:
        parts.append(0)
    return tuple(parts)


def _check_new_version(journal, current_version, startup):
    new_version = get_new_version_number()

    if new_version is not None:
        newer_version_available = _get_version_tuple(new_version) > _get_version_tuple(
            current_version
        )
    else:
        logging.error("New version info could not be read")
        new_version = _("unknown")
        newer_version_available = None

    logging.info(
        f"Current version: {current_version}, latest version: {new_version}, newer: {newer_version_available}"
    )

    if newer_version_available or not startup:
        GLib.idle_add(
            _show_update_dialog, journal, current_version, new_version, startup
        )


def check_new_version(journal, current_version, startup):
    thread = threading.Thread(
        target=_check_new_version, args=(journal, current_version, startup), daemon=True
    )
    thread.start()


def show_html_in_browser(html, filename):
    filesystem.write_file(filename, html)

    html_file = os.path.abspath(filename)
    html_file = f"file://{html_file}"
    webbrowser.open(html_file)


class StreamDuplicator:
    def __init__(self, streams):
        self.streams = streams

    def write(self, buf):
        for stream in self.streams:
            stream.write(buf)
            # If we don't flush here, stderr messages are printed late.
            stream.flush()

    def flush(self):
        for stream in self.streams:
            stream.flush()

    def close(self):
        for stream in self.streams:
            stream.close()


def safe_strxfrm(value):
    """
    Safely apply locale-aware sorting. If locale.strxfrm fails, fall back to default sorting.
    """
    try:
        return locale.strxfrm(value)
    except OSError:
        return value



# -----------------------------------------------------------------------
# Copyright (c) 2008-2024 Jendrik Seipp
#
# RedNotebook is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# RedNotebook is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with RedNotebook; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
# -----------------------------------------------------------------------

import logging
import os
import platform

from gi.repository import Gtk

from rednotebook import info
from rednotebook.configuration import Config
from rednotebook.gui import editor
from rednotebook.gui.customwidgets import ActionButton, CustomComboBoxEntry, UrlButton
from rednotebook.util import dates, filesystem, utils


class Option(Gtk.HBox):
    def __init__(self, text, option_name, tooltip=""):
        Gtk.HBox.__init__(self)

        self.text = text
        self.option_name = option_name

        self.set_spacing(5)

        self.label = Gtk.Label(label=self.text)
        self.pack_start(self.label, False, False, 0)

        if tooltip:
            self.set_tooltip_text(tooltip)

    def get_value(self):
        raise NotImplementedError


class TickOption(Option):
    def __init__(self, text, name, value=None, tooltip=""):
        Option.__init__(self, "", name, tooltip=tooltip)

        self.check_button = Gtk.CheckButton(text)
        if value is None:
            self.check_button.set_active(Option.config.read(name) == 1)
        else:
            self.check_button.set_active(value)
        self.check_button.connect("clicked", self.on_check_button_clicked)
        self.pack_start(self.check_button, False, False, 0)

    def on_check_button_clicked(self, widget):
        pass
        # TODO: Apply corresponding actions.

    def get_value(self):
        """
        We use 0 and 1 internally for bool options
        """
        return int(self.check_button.get_active())


class AutostartOption(TickOption):
    def __init__(self):
        self.autostart_file = os.path.expanduser(
            "~/.config/autostart/rednotebook.desktop"
        )
        autostart_file_exists = os.path.exists(self.autostart_file)
        TickOption.__init__(
            self, _("Load RedNotebook at startup"), None, value=autostart_file_exists
        )

    def get_value(self):
        return self.check_button.get_active()

    def set(self):
        """Apply the current setting"""
        if self.get_value():
            # Add autostart file if it is not present
            filesystem.make_file_with_dir(self.autostart_file, info.desktop_file)
        elif os.path.exists(self.autostart_file):
            os.remove(self.autostart_file)


class TextOption(Option):
    def __init__(self, text, option_name, default="", **kwargs):
        Option.__init__(self, text, option_name, **kwargs)

        # directly read the string, not the list
        value = Option.config.read(option_name, default)

        # Ensure that we have a string here
        value = str(value)

        self.entry = Gtk.Entry()
        self.entry.set_text(value)

        self.pack_start(self.entry, True, True, 0)

    def get_value(self):
        return self.entry.get_text()


class IntegerOption(Option):
    def __init__(
        self,
        text,
        option_name,
        default=0,
        min_value=0,
        max_value=1000,
        increment=1,
        **kwargs,
    ):
        Option.__init__(self, text, option_name, **kwargs)

        value = Option.config.read(option_name, default=default)
        value = int(value)

        self.spin_button = Gtk.SpinButton()
        adjustment = Gtk.Adjustment(value, min_value, max_value, increment, 10, 0)
        self.spin_button.set_adjustment(adjustment)
        self.spin_button.set_value(value)
        self.spin_button.set_numeric(True)
        self.spin_button.set_update_policy(Gtk.SpinButtonUpdatePolicy.IF_VALID)

        self.pack_start(self.spin_button, True, True, 0)

    def get_value(self):
        return self.spin_button.get_value_as_int()


class ComboBoxOption(Option):
    def __init__(self, text, name, entries, tooltip=""):
        Option.__init__(self, text, name, tooltip=tooltip)

        self.combo = CustomComboBoxEntry(Gtk.ComboBox.new_with_entry())
        self.combo.set_entries(entries)

        self.pack_start(self.combo.combo_box, False, False, 0)

    def get_value(self):
        return self.combo.get_active_text()


class DateFormatOption(ComboBoxOption):
    def __init__(self, text, name, tooltip):
        date_formats = [
            "%A, %x %X",
            _("%A, %x, Day %j"),
            "%H:%M",
            _("Week %W of Year %Y"),
            "%y-%m-%d",
            _("Day %j"),
            "%A",
            "%B",
        ]

        ComboBoxOption.__init__(self, text, name, date_formats, tooltip=tooltip)

        date_url = "http://docs.python.org/library/time.html#time.strftime"
        date_format_help_button = UrlButton(_("Help"), date_url)

        self.preview = Gtk.Label()
        self.pack_start(self.preview, False, False, 0)

        self.pack_end(date_format_help_button, False, False, 0)

        # Set default format if not present
        format = Option.config.read(name, "%A, %x %X")
        format = str(format)
        self.combo.set_active_text(format)

        self.combo.combo_box.connect("changed", self.on_format_changed)

        # Update the preview
        self.on_format_changed(None)

    def on_format_changed(self, widget):
        format_string = self.get_value()
        date_string = dates.format_date(format_string)
        # Translators: Noun
        label_text = f'{_("Preview:")} {date_string}'
        self.preview.set_text(label_text)


class FontOption(Option):
    def __init__(self, text, name):
        Option.__init__(self, text, name, "")

        self.dialog = None

        self.font_name = Option.config.read(name, editor.DEFAULT_FONT)

        self.label = Gtk.Label()
        self.label.set_text(self.font_name)

        self.button = Gtk.Button(_("Choose font ..."))
        self.button.connect("clicked", self.on_button_clicked)

        self.pack_start(self.label, False, False, 0)
        self.pack_start(self.button, False, False, 0)

    def on_button_clicked(self, widget):
        if not self.dialog:
            self.dialog = Gtk.FontSelectionDialog(_("Choose font"))

            self.dialog.set_font_name(self.font_name)
            self.dialog.set_modal(True)
            self.dialog.set_transient_for(
                Option.main_window.options_manager.dialog.dialog
            )
            self.dialog.connect("destroy", self.dialog_destroyed)
            self.dialog.get_ok_button().connect("clicked", self.font_selection_ok)
            self.dialog.get_cancel_button().connect_object(
                "clicked", lambda window: window.destroy(), self.dialog
            )

        self.dialog.show()

    def dialog_destroyed(self, widget):
        self.dialog = None

    def font_selection_ok(self, widget):
        self.font_name = self.dialog.get_font_name()
        self.label.set_text(self.font_name)
        Option.main_window.set_font(self.font_name)
        self.dialog.destroy()

    def get_value(self):
        return self.font_name


class OptionsDialog:
    def __init__(self, dialog):
        self.dialog = dialog
        self.categories = {}

    def __getattr__(self, attr):
        """Wrap the dialog"""
        return getattr(self.dialog, attr)

    def add_option(self, category, option):
        self.categories[category].pack_start(option, False, False, 0)
        option.show_all()

    def add_category(self, name, vbox):
        self.categories[name] = vbox

    def clear(self):
        for vbox in self.categories.values():
            for option in vbox.get_children():
                vbox.remove(option)


class OptionsManager:
    def __init__(self, main_window):
        self.main_window = main_window
        self.builder = main_window.builder
        self.journal = main_window.journal
        self.config = self.journal.config

        self.dialog = OptionsDialog(self.builder.get_object("options_dialog"))
        self.dialog.set_transient_for(self.main_window.main_frame)
        self.dialog.set_default_size(600, 300)
        self.dialog.add_category("general", self.builder.get_object("general_vbox"))

    def on_options_dialog(self):
        self.dialog.clear()

        # Make the config globally available
        Option.config = self.config
        Option.main_window = self.main_window

        self.options = []

        if platform.system() == "Linux" and os.path.exists("/usr/bin/rednotebook"):
            logging.debug("Running on Linux. Is installed. Adding autostart option")
            self.options.insert(0, AutostartOption())

        # Most modern Linux distributions do not have a systray anymore.
        # If this option is activated on a system without a systray, the
        # application keeps on running in the background after it has been
        # closed. The option can still be activated in the configuration file.
        if filesystem.has_system_tray():
            self.options.append(
                TickOption(
                    _("Close to system tray"),
                    "closeToTray",
                    tooltip=_("Closing the window will send RedNotebook to the tray"),
                )
            )

        # Automatic switching between preview and edit mode.
        self.options.append(
            TickOption(
                _("Switch between edit and preview mode automatically"),
                "autoSwitchMode",
            )
        )

        # Check for new version
        check_version_option = TickOption(
            _("Check for new version at startup"), "checkForNewVersion"
        )

        self.options.append(TickOption(_("Search as you type"), "instantSearch"))

        self.options.append(TickOption(_("Auto indent"), "autoIndent"))

        def check_version_action(widget):
            utils.check_new_version(
                self.main_window.journal, info.version, startup=False
            )
            # Apply changes from dialog to options window
            check = bool(self.journal.config.read("checkForNewVersion"))
            check_version_option.check_button.set_active(check)

        check_version_button = ActionButton(_("Check now"), check_version_action)
        check_version_option.pack_start(check_version_button, False, False, 0)
        self.options.append(check_version_option)

        self.options.extend(
            [
                # Use separate fonts since the preview often doesn't support the edit font.
                FontOption(_("Edit font:"), "mainFont"),
                TextOption(
                    _("Preview font:"),
                    "previewFont",
                    default=Config.defaults["previewFont"],
                    tooltip=_("Comma-separated font names"),
                ),
                DateFormatOption(
                    _("Date/Time format"),
                    "dateTimeString",
                    tooltip=_("Used by Date/Time button and $date$ template macro."),
                ),
                DateFormatOption(
                    _("Date format"),
                    "exportDateFormat",
                    tooltip=_("Used for dates in titlebar and exports."),
                ),
                IntegerOption(
                    _("Tags in cloud"),
                    "cloudMaxTags",
                    tooltip=_("Maximum number of tags displayed in the cloud"),
                ),
                TextOption(
                    _("Exclude from cloud"),
                    "cloudIgnoreList",
                    tooltip=_(
                        "Do not show these comma separated words and #tags in the clouds"
                    ),
                ),
                TextOption(
                    _("Include small words in cloud"),
                    "cloudIncludeList",
                    tooltip=_("Allow these words with 4 letters or less"),
                ),
            ]
        )

        self.add_all_options()

        response = self.dialog.run()

        if response == Gtk.ResponseType.OK:
            self.save_options()

            # Apply some options
            self.main_window.cloud.update_lists()
            self.main_window.cloud.update(force_update=True)
            self.main_window.set_auto_indent()

            visible = self.config.read("closeToTray") == 1
            # The tray icon is now only instantiated on Windows.
            if self.main_window.tray_icon:
                self.main_window.tray_icon.set_visible(visible)
        else:
            # Reset some options
            self.main_window.set_font(self.config.read("mainFont", editor.DEFAULT_FONT))

        self.dialog.hide()

    def add_all_options(self):
        for option in self.options:
            self.dialog.add_option("general", option)

    def save_options(self):
        logging.debug("Saving Options")
        for option in self.options:
            value = option.get_value()
            if option.option_name is not None:
                logging.debug(f"Setting {option.option_name} = {repr(value)}")
                self.config[option.option_name] = value
            else:
                # We don't save the autostart setting in the config file
                option.set()



# -----------------------------------------------------------------------
# Copyright (c) 2008-2024 Jendrik Seipp
#
# RedNotebook is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# RedNotebook is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with RedNotebook; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
# -----------------------------------------------------------------------

from collections import defaultdict
import datetime
import itertools
import logging
import os
import sys


def _fix_webkit_compositing_on_kde():
    """
    Fix for black word cloud and preview panes on KDE environments.

    WebKit2 has compositing issues on KDE that cause HTML content to appear black.
    Setting WEBKIT_DISABLE_COMPOSITING_MODE=1 works around this issue.

    See: https://github.com/jendrikseipp/rednotebook/issues/828
    """
    # Check if running in KDE environment
    desktop_session = os.environ.get("DESKTOP_SESSION", "").lower()
    kde_session = os.environ.get("KDE_SESSION_VERSION", "")
    xdg_desktop = os.environ.get("XDG_CURRENT_DESKTOP", "").lower()

    is_kde = (
        "kde" in desktop_session
        or "plasma" in desktop_session
        or kde_session
        or "kde" in xdg_desktop
        or "plasma" in xdg_desktop
    )

    if is_kde:
        # Logging is not yet configured at this point, so we use a plain print.
        os.environ["WEBKIT_DISABLE_COMPOSITING_MODE"] = "1"
        print(
            f"WebKit compositing disabled for KDE environment "
            f"(DESKTOP_SESSION={desktop_session}, XDG_CURRENT_DESKTOP={xdg_desktop}, "
            f"KDE_SESSION_VERSION={kde_session})."
        )


# Apply WebKit compositing fix before any GUI initialization.
_fix_webkit_compositing_on_kde()

# Use basic stdout logging before we can initialize logging correctly.
logging.basicConfig(
    level=logging.DEBUG, format="%(levelname)-8s %(message)s", stream=sys.stdout
)

try:
    import gi
except ImportError as err:
    logging.error(
        'pygobject could not be imported: "{}". Please install it (python3-gi).'.format(
            err
        )
    )
    sys.exit(1)

try:
    gi.require_version("Gtk", "3.0")
except ValueError as err:
    logging.error(err)
    sys.exit("Please install GTK (gir1.2-gtk-3.0).")

try:
    gi.require_version("GtkSource", "4")
    logging.info("Using GtkSourceView 4")
except ValueError:
    try:
        gi.require_version("GtkSource", "3.0")
        logging.info("Using GtkSourceView 3.0")
    except ValueError:
        sys.exit(
            "Please install GtkSource (gir1.2-gtksource-3.0 or gir1.2-gtksource-4)."
        )


if hasattr(sys, "frozen"):
    base_dir = sys._MEIPASS
else:
    app_dir = os.path.dirname(os.path.abspath(__file__))
    base_dir = os.path.dirname(app_dir)

print(f"Adding {base_dir} to sys.path")
sys.path.insert(0, base_dir)

from rednotebook.util import filesystem


# ---------------------- Enable i18n -------------------------------

from rednotebook.external import elibintl

LOCALE_PATH = filesystem.locale_dir
logging.info(f"Locale path: {LOCALE_PATH}")

GETTEXT_DOMAIN = "rednotebook"

"""
On Windows, translations currently only work with a hack. We need to
translate

  1) Strings in source code.
  2) GTK stock items.
  3) Strings in the Glade file.

1) works fine. For 3) we need a workaround in main_window.py. If we
initialize libintl with elibintl, the translations for 2) stop working.
Without the workaround we need to initialize libintl to avoid UTF-8
encoding errors.

When the problem is fixed upstream, we can pass libintl='libintl-8.dll'
to elibintl. See also

https://bugzilla.gnome.org/show_bug.cgi?id=574520
https://bugzilla.gnome.org/show_bug.cgi?id=753991
https://github.com/tobias47n9e/pygobject-locale
https://sourceforge.net/p/pygobjectwin32/tickets/27/

"""

elibintl.install(GETTEXT_DOMAIN, LOCALE_PATH, libintl=None)

# ------------------- end Enable i18n -------------------------------


from rednotebook.util import utils
from rednotebook.util import markup
from rednotebook.help import example_content
from rednotebook import info
from rednotebook import configuration
from rednotebook import data


args = info.get_commandline_parser().parse_args()

# ---------------------- Enable logging -------------------------------


def setup_logging(log_file):
    file_logging_stream = open(log_file, "w")

    # We want to have all stdout and stderr messages in the logfile.
    # In the frozen version we cannot log to sys.stderr because it's
    # broken on windows. Stdout doesn't work either it seems.
    stderr_streams = [file_logging_stream]
    stdout_streams = [file_logging_stream]
    if not filesystem.main_is_frozen():
        stderr_streams.append(sys.__stderr__)
        stdout_streams.append(sys.__stdout__)
    sys.stderr = utils.StreamDuplicator(stderr_streams)
    sys.stdout = utils.StreamDuplicator(stdout_streams)

    root_logger = logging.getLogger("")
    root_logger.setLevel(logging.DEBUG)

    # Python adds a default handler if some log is generated before here
    # Remove all handlers that have been added automatically
    for handler in root_logger.handlers:
        root_logger.removeHandler(handler)

    # define a Handler which writes messages to sys.stdout
    console = logging.StreamHandler(sys.stdout)
    console.setLevel(logging.DEBUG)
    # set a format which is simpler for console use
    formatter = logging.Formatter("%(asctime)s %(levelname)-8s %(message)s")
    # tell the handler to use this format
    console.setFormatter(formatter)
    # add the handler to the root logger
    root_logger.addHandler(console)

    logging.info('Writing log to file "%s"' % log_file)


default_config_file = os.path.join(filesystem.app_dir, "files", "default.cfg")
default_config = configuration.Config(default_config_file)

dirs = filesystem.Filenames(default_config)
setup_logging(dirs.log_file)

# ------------------ end Enable logging -------------------------------

logging.info("System encoding: %s" % filesystem.ENCODING)
logging.info("Language code: %s" % filesystem.LANGUAGE)

try:
    import enchant

    logging.info(f"Spell checking languages: {enchant.list_languages()}")
    logging.info(f"Spell checking dictionaries: {enchant.list_dicts()}")
except ImportError:
    pass

try:
    from gi.repository import Gtk
    from gi.repository import Gio
    from gi.repository import GLib
except (ImportError, AssertionError) as e:
    logging.error(e)
    logging.error("GTK not found. Please install it (gir1.2-gtk-3.0).")
    sys.exit(1)


from rednotebook.util import dates
from rednotebook import backup

from rednotebook.util.statistics import Statistics
from rednotebook.gui.main_window import MainWindow
from rednotebook import storage
from rednotebook.data import Month


class Journal(Gtk.Application):
    def __init__(self, *args, **kwargs):
        super().__init__(
            *args,
            application_id="app.rednotebook.RedNotebook",
            flags=Gio.ApplicationFlags.HANDLES_COMMAND_LINE,
            **kwargs,
        )
        # Let components check if the MainWindow has been created.
        self.frame = None

    def do_startup(self):
        Gtk.Application.do_startup(self)

        self.dirs = dirs

        user_config = configuration.Config(self.dirs.config_file)
        # Apply defaults where no custom values have been set
        for key, value in default_config.items():
            if key not in user_config:
                user_config[key] = value
        self.config = user_config
        self.config.save_state()

        logging.info("Running in portable mode: %s" % self.dirs.portable)

        self.month = None
        self.date = None
        self.months = {}

        # The dir name is the title
        self.title = ""

        # show instructions at first start
        self.is_first_start = self.config.read("firstStart")
        self.config["firstStart"] = 0
        logging.info("First Start: %s" % bool(self.is_first_start))

        logging.info("RedNotebook version: %s" % info.version)
        logging.info(filesystem.get_platform_info())

        self.actual_date = self.get_start_date()

        self.do_activate()

        journal_path = self.get_journal_path()
        if not self.dirs.is_valid_journal_path(journal_path):
            logging.error(
                "Invalid directory: %s. Using default journal." % journal_path
            )
            self.show_message(
                _("You cannot use this directory for your journal:")
                + " %s" % journal_path
                + ". "
                + _("Opening default journal."),
                error=True,
            )
            journal_path = self.dirs.default_data_dir
        self.open_journal(journal_path)

        self.archiver = backup.Archiver(self)
        GLib.idle_add(self.archiver.check_last_backup_date)

        # Check for a new version
        if self.config.read("checkForNewVersion") == 1:
            utils.check_new_version(self, info.version, startup=True)

        # Automatically save the content after a period of time
        GLib.timeout_add_seconds(600, self.save_to_disk)

    def do_activate(self):
        if not self.frame:
            self.frame = MainWindow(self)
        self.frame.main_frame.present()

    def do_command_line(self, _command_line):
        # Arguments are parsed elsewhere, so we only show the window here.
        self.activate()
        return 0  # Must return a number.

    def get_journal_path(self):
        """
        Retrieve the path from optional args or return standard value if args
        not present
        """
        if not args.journal:
            data_dir = self.config.read("dataDir", self.dirs.data_dir)
            if not os.path.isabs(data_dir):
                data_dir = os.path.join(self.dirs.app_dir, data_dir)
                data_dir = os.path.normpath(data_dir)
            return data_dir

        # path_arg can be e.g. data (under .rednotebook), data (elsewhere),
        # or an absolute path /home/username/myjournal
        # Try to find the journal under the standard location or at the given
        # absolute or relative location
        path_arg = args.journal

        logging.debug('Trying to find journal "%s"' % path_arg)

        paths_to_check = [path_arg, os.path.join(self.dirs.journal_user_dir, path_arg)]

        for path in paths_to_check:
            if os.path.exists(path):
                if os.path.isdir(path):
                    return path
                else:
                    logging.warning(
                        "To open a journal you must specify a " "directory, not a file."
                    )

        logging.error(
            'The path "%s" is not a valid journal directory. '
            'Execute "rednotebook -h" for instructions' % path_arg
        )
        sys.exit(2)

    def get_start_date(self):
        """
        Retrieve the date from optional args or otherwise return 'today'
        """
        if not args.start_date:
            return datetime.date.today()

        try:
            return dates.get_date_from_date_string(args.start_date)
        except ValueError:
            logging.error(
                "Invalid date: %s (required format: YYYY-MM-DD)." % args.start_date
            )
            sys.exit(2)

    def exit(self):
        self.frame.add_values_to_config()

        # Make it possible to stop the program from exiting
        # e.g. if the journal could not be saved
        self.is_allowed_to_exit = True
        self.save_to_disk(exit_imminent=True)

        if self.is_allowed_to_exit:
            logging.info("Goodbye!")
            # Informs the logging system to perform an orderly shutdown by
            # flushing and closing all handlers.
            logging.shutdown()
            self.quit()

    def convert(self, text, target, headers=None, options=None, use_gtk_theme=False):
        options = options or {}
        options["font"] = self.config.read("previewFont")
        if use_gtk_theme:
            bgcolor, fgcolor = utils.get_gtk_colors(
                self.frame.day_text_field.day_text_view
            )
            options["bgcolor"] = bgcolor
            options["fgcolor"] = fgcolor
        return markup.convert(
            text, target, self.dirs.data_dir, headers=headers, options=options
        )

    def save_to_disk(self, exit_imminent=False, changing_journal=False, saveas=False):
        self.save_old_day()

        try:
            filesystem.make_directory(self.dirs.data_dir)
        except OSError as err:
            logging.error(f"Creating journal directory failed: {err}")
            self.frame.show_save_error_dialog(exit_imminent)
            return True

        try:
            something_saved = storage.save_months_to_disk(
                self.months, self.dirs.data_dir, exit_imminent, saveas
            )
        except OSError as err:
            logging.error(f"Saving month files failed: {err}")
            self.frame.show_save_error_dialog(exit_imminent)
            something_saved = None

        if something_saved:
            self.show_message(
                _("The content has been saved to %s") % self.dirs.data_dir, error=False
            )
            logging.info("The content has been saved to %r" % self.dirs.data_dir)
        elif something_saved is None:
            # Don't display this as an error, because we already show a dialog.
            self.show_message(_("The journal could not be saved"), error=False)
        else:
            self.show_message(_("Nothing to save"), error=False)

        self.config.save_to_disk()

        if not (exit_imminent or changing_journal) and something_saved:
            # Update cloud.
            self.frame.cloud.update(force_update=True)
            for tag in self.get_escaped_tags():
                self.frame.search_box.add_entry(tag)

        # tell gobject to keep saving the content in regular intervals
        return True

    def open_journal(self, data_dir):
        if not os.path.exists(data_dir):
            logging.warning(
                "The dir %s does not exist. Select a different dir." % data_dir
            )
            return

        if self.months:
            self.save_to_disk(changing_journal=True)

        logging.info("Opening journal at %r" % data_dir)
        self.dirs.data_dir = data_dir

        self.month = None
        self.months.clear()
        self.frame.search_box.clear()
        self.frame.day_text_field.clear_buffers()

        self.months = storage.load_all_months_from_disk(data_dir)

        # Nothing to save before first day change
        self.load_day(self.actual_date)

        if self.is_first_start and not os.listdir(data_dir) and not self.days:
            self.add_instruction_content()

        self.stats = Statistics(self)

        self.frame.cloud.update(force_update=True)

        self.frame.categories_tree_view.categories = self.categories
        self.frame.search_box.set_entries(self.get_escaped_tags())

        self.title = filesystem.get_journal_title(data_dir)

        # Set frame title
        self.set_frame_title()

        # Save the folder for next start
        if not self.dirs.portable:
            self.config["dataDir"] = data_dir
        else:
            rel_data_dir = filesystem.get_relative_path(self.dirs.app_dir, data_dir)
            self.config["dataDir"] = rel_data_dir

    def set_frame_title(self):
        parts = ["RedNotebook"]
        if self.title != "data":
            parts.append(self.title)
        parts.append(dates.format_date(self.config.read("exportDateFormat"), self.date))
        self.frame.main_frame.set_title(" - ".join(parts))

    def get_month(self, date):
        """
        Returns the corresponding month if it has previously been visited,
        otherwise a new month is created and returned
        """

        year_and_month = dates.get_year_and_month_from_date(date)

        # Selected month has not been loaded or created yet
        if year_and_month not in self.months:
            self.months[year_and_month] = Month(date.year, date.month)

        return self.months[year_and_month]

    def get_day(self, date):
        return self.get_month(date).get_day(date.day)

    def get_escaped_tags(self):
        return ["#%s" % data.escape_tag(tag) for tag in self.categories]

    def save_old_day(self):
        """Order is important"""
        old_content = self.day.content
        new_content = self.frame.categories_tree_view.get_day_content()
        new_content["text"] = self.frame.get_day_text()
        self.day.content = new_content

        content_changed = old_content != new_content
        if content_changed:
            self.month.edited = True

        self.frame.calendar.set_day_edited(self.date.day, not self.day.empty)

    def load_day(self, new_date):
        old_date = self.date
        self.date = new_date

        if not dates.same_month(new_date, old_date) or self.month is None:
            self.month = self.get_month(self.date)

        self.frame.set_date(self.month, self.date, self.day)

        self.set_frame_title()

    @property
    def day(self):
        return self.month.get_day(self.date.day)

    def change_date(self, new_date):
        if new_date < datetime.date(1900, 1, 1):
            self.show_message(
                "Only dates after 1900 are supported.", title="Too Early", error=True
            )
            return

        if new_date == self.date:
            return

        self.save_old_day()
        self.load_day(new_date)

    def go_to_next_day(self):
        next_date = self.date + dates.one_day
        following_edited_days = self.get_days_in_date_range(start_date=next_date)
        if following_edited_days:
            next_date = following_edited_days[0].date
        self.change_date(next_date)

    def go_to_prev_day(self):
        prev_date = self.date - dates.one_day
        previous_edited_days = self.get_days_in_date_range(end_date=prev_date)
        if previous_edited_days:
            prev_date = previous_edited_days[-1].date
        self.change_date(prev_date)

    def show_message(self, msg, title=None, error=False):
        if error and not title:
            title = _("Error")

        if error:
            msg_type = Gtk.MessageType.ERROR
            log_level = logging.ERROR
        else:
            msg_type = Gtk.MessageType.INFO
            log_level = logging.INFO

        self.frame.show_message(title, msg, msg_type)
        logging.log(log_level, f"{title}. {msg}" if title else msg)

    @property
    def categories(self):
        return sorted(
            set(itertools.chain.from_iterable(day.categories for day in self.days)),
            key=utils.safe_strxfrm,
        )

    def get_entries(self, category):
        entries = set()
        for day in self.days:
            entries |= set(day.get_entries(category))
        return sorted(entries)

    def search(self, text, tags):
        results = []
        for day in reversed(self.get_days_with_tags(tags)):
            results.append(day.search(text, tags))
        return results

    def replace_all(self, old, new):
        """Replace all strings matching 'old' with the 'new' string."""
        total_replacements = 0

        for day in self.days:
            replacements = day.replace_all(old, new)
            if replacements > 0:
                day.month.edited = True
                total_replacements += replacements

        # Update current frame.
        self.frame.set_day_text(self.day.text)

        # Update the preview if it's active.
        if self.frame.preview_mode and self.frame.html_editor.internal:
            self.frame.html_editor.show_day(self.day)

        if total_replacements > 0:
            # Prevent saving data from the buffers instead of the replaced text.
            self.frame.day_text_field.clear_buffers()

            self.frame.cloud.update(force_update=True)
            self.frame.search_box.update_search_results()

            self.show_message(
                _(f"Replaced {total_replacements} occurrences of the search text"),
                error=False,
            )
        else:
            self.show_message(_("No text has been replaced"), error=False)

    def get_days_with_tags(self, tags):
        if not tags:
            return self.days
        days = []
        for day in self.days:
            day_tags = {data.escape_tag(tag) for tag in day.categories}
            if all(tag in day_tags for tag in tags):
                days.append(day)
        return days

    def get_word_count_dict(self):
        """
        Return a dictionary mapping the words to their number of appearance.
        """
        word_dict = defaultdict(int)
        for day in self.days:
            words = day.get_words()
            for word in words:
                word_dict[word.lower()] += 1
        return word_dict

    @property
    def days(self):
        """
        Returns all edited days ordered by their date
        """
        # The day being edited counts too
        if self.frame:
            self.save_old_day()

        days = []
        for month in self.months.values():
            # Filter out days without content.
            days_in_month = [day for day in month.days.values() if not day.empty]
            days.extend(days_in_month)

        # Sort days
        days = sorted(days, key=lambda day: day.date)
        return days

    def get_days_in_date_range(self, start_date=None, end_date=None):
        if not start_date:
            start_date = datetime.date.min
        if not end_date:
            end_date = datetime.date.max

        start_date, end_date = sorted([start_date, end_date])
        assert start_date <= end_date

        days_in_date_range = []
        for day in self.days:
            if day.date < start_date:
                continue
            elif start_date <= day.date <= end_date:
                days_in_date_range.append(day)
            elif day.date > end_date:
                break
        return days_in_date_range

    def add_instruction_content(self):
        self.change_date(datetime.date.today())
        current_date = self.date

        logging.info("Adding example content on %s" % current_date)

        for example_day in example_content:
            self.day.content = example_day
            self.frame.set_date(self.month, self.date, self.day)
            self.go_to_next_day()

        self.change_date(current_date)


def main():
    journal = Journal()
    utils.setup_signal_handlers(journal)
    journal.run(sys.argv)

    try:
        logging.info(f"Peak memory: {filesystem.get_peak_memory_in_kb()} KiB")
    except Warning:
        pass


if __name__ == "__main__":
    main()
